from copy import deepcopy
from time import time
from unittest import IsolatedAsyncioTestCase
from unittest.mock import ANY

from marshmallow import ValidationError

from .....admin.request_context import AdminRequestContext
from .....anoncreds.models.presentation_request import (
    AnonCredsPresentationReqAttrSpecSchema,
)
from .....indy.holder import IndyHolder
from .....indy.verifier import IndyVerifier
from .....ledger.base import BaseLedger
from .....storage.error import StorageNotFoundError
from .....storage.vc_holder.base import VCHolder
from .....storage.vc_holder.vc_record import VCRecord
from .....tests import mock
from .....utils.testing import create_test_profile
from ...dif.pres_exch import SchemaInputDescriptor
from .. import routes as test_module
from ..messages.pres_format import V20PresFormat
from ..models.pres_exchange import V20PresExRecord

ISSUER_DID = "NcYxiDXkpYi6ov5FcYDi1e"
S_ID = f"{ISSUER_DID}:2:vidya:1.0"
CD_ID = f"{ISSUER_DID}:3:CL:{S_ID}:tag1"
RR_ID = f"{ISSUER_DID}:4:{CD_ID}:CL_ACCUM:0"
PROOF_REQ_NAME = "name"
PROOF_REQ_VERSION = "1.0"
PROOF_REQ_NONCE = "12345"

NOW = int(time())
INDY_PROOF_REQ = {
    "name": PROOF_REQ_NAME,
    "version": PROOF_REQ_VERSION,
    "nonce": PROOF_REQ_NONCE,
    "requested_attributes": {
        "0_player_uuid": {
            "name": "player",
            "restrictions": [{"cred_def_id": CD_ID}],
            "non_revoked": {"from": NOW, "to": NOW},
        },
        "1_screencapture_uuid": {
            "name": "screenCapture",
            "restrictions": [{"cred_def_id": CD_ID}],
            "non_revoked": {"from": NOW, "to": NOW},
        },
    },
    "requested_predicates": {
        "0_highscore_GE_uuid": {
            "name": "highScore",
            "p_type": ">=",
            "p_value": 1000000,
            "restrictions": [{"cred_def_id": CD_ID}],
            "non_revoked": {"from": NOW, "to": NOW},
        }
    },
}

DIF_PROOF_REQ = {
    "options": {
        "challenge": "3fa85f64-5717-4562-b3fc-2c963f66afa7",
        "domain": "4jt78h47fh47",
    },
    "presentation_definition": {
        "id": "32f54163-7166-48f1-93d8-ff217bdb0654",
        "submission_requirements": [
            {
                "name": "Citizenship Information",
                "rule": "pick",
                "min": 1,
                "from": "A",
            }
        ],
        "input_descriptors": [
            {
                "id": "citizenship_input_1",
                "name": "EU Driver's License",
                "group": ["A"],
                "schema": [
                    {"uri": "https://www.w3.org/2018/credentials#VerifiableCredential"},
                    {"uri": "https://w3id.org/citizenship#PermanentResidentCard"},
                ],
                "constraints": {
                    "limit_disclosure": "required",
                    "fields": [
                        {
                            "path": ["$.credentialSubject.givenName"],
                            "purpose": "The claim must be from one of the specified issuers",
                            "filter": {
                                "type": "string",
                                "enum": ["JOHN", "CAI"],
                            },
                        }
                    ],
                },
            }
        ],
    },
}


DIF_PRES_PROPOSAL = {
    "input_descriptors": [
        {
            "id": "citizenship_input_1",
            "name": "EU Driver's License",
            "group": ["A"],
            "schema": [
                {"uri": "https://www.w3.org/2018/credentials#VerifiableCredential"},
                {"uri": "https://w3id.org/citizenship#PermanentResidentCard"},
            ],
            "constraints": {
                "limit_disclosure": "required",
                "fields": [
                    {
                        "path": ["$.credentialSubject.givenName"],
                        "purpose": "The claim must be from one of the specified issuers",
                        "filter": {"type": "string", "enum": ["JOHN", "CAI"]},
                    }
                ],
            },
        }
    ]
}


class TestPresentProofRoutes(IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        self.profile = await create_test_profile(
            settings={
                "admin.admin_api_key": "secret-key",
            }
        )
        self.context = AdminRequestContext.test_context({}, profile=self.profile)
        injector = self.profile.context.injector

        self.ledger = mock.MagicMock(BaseLedger, autospec=True)
        self.ledger.get_schema = mock.CoroutineMock(return_value=mock.MagicMock())
        self.ledger.get_credential_definition = mock.CoroutineMock(
            return_value={"value": {"revocation": {"...": "..."}}}
        )
        self.ledger.get_revoc_reg_def = mock.CoroutineMock(
            return_value={
                "ver": "1.0",
                "id": RR_ID,
                "revocDefType": "CL_ACCUM",
                "tag": RR_ID.split(":")[-1],
                "credDefId": CD_ID,
                "value": {
                    "IssuanceType": "ISSUANCE_BY_DEFAULT",
                    "maxCredNum": 1000,
                    "publicKeys": {"accumKey": {"z": "1 ..."}},
                    "tailsHash": "3MLjUFQz9x9n5u9rFu8Ba9C5bo4HNFjkPNc54jZPSNaZ",
                    "tailsLocation": "http://sample.ca/path",
                },
            }
        )
        self.ledger.get_revoc_reg_delta = mock.CoroutineMock(
            return_value=(
                {
                    "ver": "1.0",
                    "value": {"prevAccum": "1 ...", "accum": "21 ...", "issued": [1]},
                },
                NOW,
            )
        )
        self.ledger.get_revoc_reg_entry = mock.CoroutineMock(
            return_value=(
                {
                    "ver": "1.0",
                    "value": {"prevAccum": "1 ...", "accum": "21 ...", "issued": [1]},
                },
                NOW,
            )
        )
        injector.bind_instance(BaseLedger, self.ledger)

        self.request_dict = {
            "context": self.context,
            "outbound_message_router": mock.CoroutineMock(),
        }
        self.request = mock.MagicMock(
            app={},
            match_info={},
            query={},
            __getitem__=lambda _, k: self.request_dict[k],
            headers={"x-api-key": "secret-key"},
        )

    async def test_validate(self):
        schema = test_module.V20PresProposalByFormatSchema()
        schema.validate_fields({"indy": {"attributes": [], "predicates": []}})
        schema.validate_fields({"dif": {"some_dif_criterion": "..."}})
        schema.validate_fields(
            {
                "indy": {"attributes": [], "predicates": []},
                "dif": {"some_dif_criterion": "..."},
            }
        )
        with self.assertRaises(test_module.ValidationError):
            schema.validate_fields({})
        with self.assertRaises(test_module.ValidationError):
            schema.validate_fields({"veres-one": {"no": "support"}})

        schema = test_module.V20PresRequestByFormatSchema()
        schema.validate_fields({"indy": {"...": "..."}})
        schema.validate_fields({"dif": {"...": "..."}})
        schema.validate_fields({"indy": {"...": "..."}, "dif": {"...": "..."}})
        with self.assertRaises(test_module.ValidationError):
            schema.validate_fields({})
        with self.assertRaises(test_module.ValidationError):
            schema.validate_fields({"veres-one": {"no": "support"}})

        schema = test_module.V20PresSpecByFormatRequestSchema()
        schema.validate_fields({"indy": {"...": "..."}})
        schema.validate_fields({"dif": {"...": "..."}})
        schema.validate_fields({"indy": {"...": "..."}, "dif": {"...": "..."}})
        with self.assertRaises(test_module.ValidationError):
            schema.validate_fields({})
        with self.assertRaises(test_module.ValidationError):
            schema.validate_fields({"veres-one": {"no": "support"}})

    async def test_validate_proof_req_attr_spec(self):
        aspec = AnonCredsPresentationReqAttrSpecSchema()
        aspec.validate_fields({"name": "attr0"})
        aspec.validate_fields(
            {
                "names": ["attr0", "attr1"],
                "restrictions": [{"attr::attr1::value": "my-value"}],
            }
        )
        aspec.validate_fields(
            {"name": "attr0", "restrictions": [{"schema_name": "preferences"}]}
        )
        with self.assertRaises(ValidationError):
            aspec.validate_fields({})
        with self.assertRaises(ValidationError):
            aspec.validate_fields({"name": "attr0", "names": ["attr1", "attr2"]})
        with self.assertRaises(ValidationError):
            aspec.validate_fields({"names": ["attr1", "attr2"]})
        with self.assertRaises(ValidationError):
            aspec.validate_fields({"names": ["attr0", "attr1"], "restrictions": []})
        with self.assertRaises(ValidationError):
            aspec.validate_fields({"names": ["attr0", "attr1"], "restrictions": [{}]})

    async def test_present_proof_list(self):
        self.request.query = {
            "thread_id": "thread_id_0",
            "connection_id": "conn_id_0",
            "role": "dummy",
            "state": "dummy",
        }

        mock_pres_ex_rec_inst = mock.MagicMock(
            serialize=mock.MagicMock(return_value={"thread_id": "sample-thread-id"})
        )
        with (
            mock.patch.object(
                test_module, "V20PresExRecord", autospec=True
            ) as mock_pres_ex_rec_cls,
            mock.patch.object(
                test_module.web, "json_response", mock.MagicMock()
            ) as mock_response,
        ):
            mock_pres_ex_rec_cls.query = mock.CoroutineMock(
                return_value=[mock_pres_ex_rec_inst]
            )

            await test_module.present_proof_list(self.request)
            mock_response.assert_called_once_with(
                {"results": [mock_pres_ex_rec_inst.serialize.return_value]}
            )

    async def test_present_proof_list_x(self):
        self.request.query = {
            "thread_id": "thread_id_0",
            "connection_id": "conn_id_0",
            "role": "dummy",
            "state": "dummy",
        }

        with mock.patch.object(
            test_module, "V20PresExRecord", autospec=True
        ) as mock_pres_ex_rec_cls:
            mock_pres_ex_rec_cls.query = mock.CoroutineMock(
                side_effect=test_module.StorageError()
            )

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.present_proof_list(self.request)

    async def test_present_proof_credentials_list_not_found(self):
        self.request.match_info = {"pres_ex_id": "dummy"}

        with mock.patch.object(
            test_module, "V20PresExRecord", autospec=True
        ) as mock_pres_ex_rec_cls:
            mock_pres_ex_rec_cls.retrieve_by_id = mock.CoroutineMock()

            # Emulate storage not found (bad presentation exchange id)
            mock_pres_ex_rec_cls.retrieve_by_id.side_effect = StorageNotFoundError()

            with self.assertRaises(test_module.web.HTTPNotFound):
                await test_module.present_proof_credentials_list(self.request)

    async def test_present_proof_credentials_x(self):
        self.request.match_info = {
            "pres_ex_id": "123-456-789",
            "referent": "myReferent1",
        }
        self.request.query = {"extra_query": {}}
        mock_holder = mock.MagicMock(IndyHolder, autospec=True)
        mock_holder.get_credentials_for_presentation_request_by_referent = (
            mock.CoroutineMock(side_effect=test_module.IndyHolderError())
        )
        self.profile.context.injector.bind_instance(IndyHolder, mock_holder)
        mock_px_rec = mock.MagicMock(save_error_state=mock.CoroutineMock())

        with mock.patch.object(
            test_module, "V20PresExRecord", autospec=True
        ) as mock_pres_ex_rec_cls:
            mock_pres_ex_rec_cls.retrieve_by_id = mock.CoroutineMock(
                return_value=mock_px_rec
            )

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.present_proof_credentials_list(self.request)

    async def test_present_proof_credentials_list_single_referent(self):
        self.request.match_info = {
            "pres_ex_id": "123-456-789",
            "referent": "myReferent1",
        }
        self.request.query = {"extra_query": {}}

        returned_credentials = [{"name": "Credential1"}, {"name": "Credential2"}]
        mock_holder = mock.MagicMock(IndyHolder, autospec=True)
        mock_holder.get_credentials_for_presentation_request_by_referent = (
            mock.CoroutineMock(return_value=returned_credentials)
        )
        self.profile.context.injector.bind_instance(IndyHolder, mock_holder)

        with (
            mock.patch.object(
                test_module, "V20PresExRecord", autospec=True
            ) as mock_pres_ex_rec_cls,
            mock.patch.object(
                test_module.web, "json_response", mock.MagicMock()
            ) as mock_response,
        ):
            mock_pres_ex_rec_cls.return_value = mock.MagicMock(
                retrieve_by_id=mock.CoroutineMock()
            )

            await test_module.present_proof_credentials_list(self.request)
            mock_response.assert_called_once_with(returned_credentials)

    async def test_present_proof_credentials_list_multiple_referents(self):
        self.request.match_info = {
            "pres_ex_id": "123-456-789",
            "referent": "myReferent1,myReferent2",
        }
        self.request.query = {"extra_query": {}}

        returned_credentials = [{"name": "Credential1"}, {"name": "Credential2"}]
        mock_holder = mock.MagicMock(IndyHolder, autospec=True)
        mock_holder.get_credentials_for_presentation_request_by_referent = (
            mock.CoroutineMock(return_value=returned_credentials)
        )
        self.profile.context.injector.bind_instance(IndyHolder, mock_holder)

        with (
            mock.patch.object(
                test_module, "V20PresExRecord", autospec=True
            ) as mock_pres_ex_rec_cls,
            mock.patch.object(
                test_module.web, "json_response", mock.MagicMock()
            ) as mock_response,
        ):
            mock_pres_ex_rec_cls.return_value = mock.MagicMock(
                retrieve_by_id=mock.CoroutineMock()
            )

            await test_module.present_proof_credentials_list(self.request)
            mock_response.assert_called_once_with(returned_credentials)

    async def test_present_proof_credentials_list_dif(self):
        self.request.match_info = {
            "pres_ex_id": "123-456-789",
        }
        self.request.query = {"extra_query": {}}

        returned_credentials = [
            mock.MagicMock(cred_value={"name": "Credential1"}),
            mock.MagicMock(cred_value={"name": "Credential2"}),
        ]
        mock_indy_holder = mock.MagicMock(IndyHolder, autospec=True)
        mock_indy_holder.get_credentials_for_presentation_request_by_referent = (
            mock.CoroutineMock(return_value=returned_credentials)
        )
        self.profile.context.injector.bind_instance(IndyHolder, mock_indy_holder)

        mock_vc_holder = mock.MagicMock(VCHolder, autospec=True)
        mock_vc_holder.search_credentials = mock.MagicMock(
            return_value=mock.MagicMock(
                fetch=mock.CoroutineMock(return_value=returned_credentials)
            )
        )
        self.profile.context.injector.bind_instance(VCHolder, mock_vc_holder)
        record = V20PresExRecord(
            state="request-received",
            role="prover",
            pres_proposal=None,
            pres_request={
                "@type": "https://didcomm.org/present-proof/2.0/request-presentation",
                "@id": "6ae00c6c-87fa-495a-b546-5f5953817c92",
                "comment": "string",
                "formats": [
                    {
                        "attach_id": "dif",
                        "format": "dif/presentation-exchange/definitions@v1.0",
                    }
                ],
                "request_presentations~attach": [
                    {
                        "@id": "dif",
                        "mime-type": "application/json",
                        "data": {"json": DIF_PROOF_REQ},
                    }
                ],
                "will_confirm": True,
            },
            pres=None,
            verified=None,
            auto_present=False,
            error_msg=None,
        )

        with (
            mock.patch.object(
                test_module, "V20PresExRecord", autospec=True
            ) as mock_pres_ex_rec_cls,
            mock.patch.object(
                test_module.web, "json_response", mock.MagicMock()
            ) as mock_response,
        ):
            mock_pres_ex_rec_cls.retrieve_by_id.return_value = record

            await test_module.present_proof_credentials_list(self.request)
            mock_response.assert_called_once_with(
                [
                    {"name": "Credential1", "record_id": ANY},
                    {"name": "Credential2", "record_id": ANY},
                ]
            )

    async def test_present_proof_credentials_list_dif_one_of_filter(self):
        self.request.match_info = {
            "pres_ex_id": "123-456-789",
        }
        self.request.query = {"extra_query": {}}

        returned_credentials = [
            mock.MagicMock(cred_value={"name": "Credential1"}, record_id="test_1"),
            mock.MagicMock(cred_value={"name": "Credential2"}, record_id="test_2"),
        ]
        mock_indy_holder = mock.MagicMock(IndyHolder, autospec=True)
        mock_indy_holder.get_credentials_for_presentation_request_by_referent = (
            mock.CoroutineMock(return_value=returned_credentials)
        )
        self.profile.context.injector.bind_instance(IndyHolder, mock_indy_holder)

        mock_vc_holder = mock.MagicMock(VCHolder, autospec=True)
        mock_vc_holder.search_credentials = mock.MagicMock(
            return_value=mock.MagicMock(
                fetch=mock.CoroutineMock(return_value=returned_credentials)
            )
        )
        self.profile.context.injector.bind_instance(VCHolder, mock_vc_holder)
        pres_request = deepcopy(DIF_PROOF_REQ)
        pres_request["presentation_definition"]["input_descriptors"][0]["schema"] = {
            "oneof_filter": [
                [
                    {"uri": "https://www.w3.org/2018/credentials#VerifiableCredential"},
                    {"uri": "https://w3id.org/citizenship#PermanentResidentCard"},
                ],
                [{"uri": "https://www.w3.org/Test#Test"}],
            ]
        }
        record = V20PresExRecord(
            state="request-received",
            role="prover",
            pres_proposal=None,
            pres_request={
                "@type": "https://didcomm.org/present-proof/2.0/request-presentation",
                "@id": "6ae00c6c-87fa-495a-b546-5f5953817c92",
                "comment": "string",
                "formats": [
                    {
                        "attach_id": "dif",
                        "format": "dif/presentation-exchange/definitions@v1.0",
                    }
                ],
                "request_presentations~attach": [
                    {
                        "@id": "dif",
                        "mime-type": "application/json",
                        "data": {"json": pres_request},
                    }
                ],
                "will_confirm": True,
            },
            pres=None,
            verified=None,
            auto_present=False,
            error_msg=None,
        )

        with (
            mock.patch.object(
                test_module, "V20PresExRecord", autospec=True
            ) as mock_pres_ex_rec_cls,
            mock.patch.object(
                test_module.web, "json_response", mock.MagicMock()
            ) as mock_response,
        ):
            mock_pres_ex_rec_cls.retrieve_by_id.return_value = record

            await test_module.present_proof_credentials_list(self.request)
            mock_response.assert_called_once_with(
                [
                    {"name": "Credential1", "record_id": "test_1"},
                    {"name": "Credential2", "record_id": "test_2"},
                ]
            )

    async def test_present_proof_credentials_dif_no_tag_query(self):
        self.request.match_info = {
            "pres_ex_id": "123-456-789",
        }
        self.request.query = {"extra_query": {}}
        test_pd = deepcopy(DIF_PROOF_REQ)
        test_pd["presentation_definition"]["input_descriptors"][0]["schema"][0][
            "required"
        ] = False
        test_pd["presentation_definition"]["input_descriptors"][0]["schema"][1][
            "required"
        ] = False
        returned_credentials = [
            mock.MagicMock(cred_value={"name": "Credential1"}),
            mock.MagicMock(cred_value={"name": "Credential2"}),
        ]
        mock_indy_holder = mock.MagicMock(IndyHolder, autospec=True)
        mock_indy_holder.get_credentials_for_presentation_request_by_referent = (
            mock.CoroutineMock(return_value=returned_credentials)
        )
        self.profile.context.injector.bind_instance(IndyHolder, mock_indy_holder)

        mock_vc_holder = mock.MagicMock(VCHolder, autospec=True)
        mock_vc_holder.search_credentials = mock.MagicMock(
            return_value=mock.MagicMock(
                fetch=mock.CoroutineMock(return_value=returned_credentials)
            )
        )
        self.profile.context.injector.bind_instance(VCHolder, mock_vc_holder)
        record = V20PresExRecord(
            state="request-received",
            role="prover",
            pres_proposal=None,
            pres_request={
                "@type": "https://didcomm.org/present-proof/2.0/request-presentation",
                "@id": "6ae00c6c-87fa-495a-b546-5f5953817c92",
                "comment": "string",
                "formats": [
                    {
                        "attach_id": "dif",
                        "format": "dif/presentation-exchange/definitions@v1.0",
                    }
                ],
                "request_presentations~attach": [
                    {
                        "@id": "dif",
                        "mime-type": "application/json",
                        "data": {"json": test_pd},
                    }
                ],
                "will_confirm": True,
            },
            pres=None,
            verified=None,
            auto_present=False,
            error_msg=None,
        )

        with (
            mock.patch.object(
                test_module, "V20PresExRecord", autospec=True
            ) as mock_pres_ex_rec_cls,
            mock.patch.object(
                test_module.web, "json_response", mock.MagicMock()
            ) as mock_response,
        ):
            mock_pres_ex_rec_cls.retrieve_by_id.return_value = record

            await test_module.present_proof_credentials_list(self.request)
            mock_response.assert_called_once_with(
                [
                    {"name": "Credential1", "record_id": ANY},
                    {"name": "Credential2", "record_id": ANY},
                ]
            )

    async def test_present_proof_credentials_single_ldp_vp_claim_format(self):
        self.request.match_info = {
            "pres_ex_id": "123-456-789",
        }
        self.request.query = {"extra_query": {}}
        test_pd = deepcopy(DIF_PROOF_REQ)
        test_pd["presentation_definition"]["format"] = {
            "ldp_vp": {"proof_type": ["Ed25519Signature2018"]}
        }
        del test_pd["presentation_definition"]["input_descriptors"][0]["constraints"][
            "limit_disclosure"
        ]
        returned_credentials = [
            mock.MagicMock(cred_value={"name": "Credential1"}),
            mock.MagicMock(cred_value={"name": "Credential2"}),
        ]
        mock_indy_holder = mock.MagicMock(IndyHolder, autospec=True)
        mock_indy_holder.get_credentials_for_presentation_request_by_referent = (
            mock.CoroutineMock(return_value=returned_credentials)
        )
        self.profile.context.injector.bind_instance(IndyHolder, mock_indy_holder)

        mock_vc_holder = mock.MagicMock(VCHolder, autospec=True)
        mock_vc_holder.search_credentials = mock.MagicMock(
            return_value=mock.MagicMock(
                fetch=mock.CoroutineMock(return_value=returned_credentials)
            )
        )
        self.profile.context.injector.bind_instance(VCHolder, mock_vc_holder)
        record = V20PresExRecord(
            state="request-received",
            role="prover",
            pres_proposal=None,
            pres_request={
                "@type": "https://didcomm.org/present-proof/2.0/request-presentation",
                "@id": "6ae00c6c-87fa-495a-b546-5f5953817c92",
                "comment": "string",
                "formats": [
                    {
                        "attach_id": "dif",
                        "format": "dif/presentation-exchange/definitions@v1.0",
                    }
                ],
                "request_presentations~attach": [
                    {
                        "@id": "dif",
                        "mime-type": "application/json",
                        "data": {"json": test_pd},
                    }
                ],
                "will_confirm": True,
            },
            pres=None,
            verified=None,
            auto_present=False,
            error_msg=None,
        )

        with (
            mock.patch.object(
                test_module, "V20PresExRecord", autospec=True
            ) as mock_pres_ex_rec_cls,
            mock.patch.object(
                test_module.web, "json_response", mock.MagicMock()
            ) as mock_response,
        ):
            mock_pres_ex_rec_cls.retrieve_by_id.return_value = record

            await test_module.present_proof_credentials_list(self.request)
            mock_response.assert_called_once_with(
                [
                    {"name": "Credential1", "record_id": ANY},
                    {"name": "Credential2", "record_id": ANY},
                ]
            )

    async def test_present_proof_credentials_double_ldp_vp_claim_format(self):
        self.request.match_info = {
            "pres_ex_id": "123-456-789",
        }
        self.request.query = {"extra_query": {}}
        test_pd = deepcopy(DIF_PROOF_REQ)
        test_pd["presentation_definition"]["format"] = {
            "ldp_vp": {"proof_type": ["BbsBlsSignature2020", "Ed25519Signature2018"]}
        }
        del test_pd["presentation_definition"]["input_descriptors"][0]["constraints"][
            "limit_disclosure"
        ]
        returned_credentials = [
            mock.MagicMock(cred_value={"name": "Credential1"}),
            mock.MagicMock(cred_value={"name": "Credential2"}),
        ]
        mock_indy_holder = mock.MagicMock(IndyHolder, autospec=True)
        mock_indy_holder.get_credentials_for_presentation_request_by_referent = (
            mock.CoroutineMock(return_value=returned_credentials)
        )
        self.profile.context.injector.bind_instance(IndyHolder, mock_indy_holder)

        mock_vc_holder = mock.MagicMock(VCHolder, autospec=True)
        mock_vc_holder.search_credentials = mock.MagicMock(
            return_value=mock.MagicMock(
                fetch=mock.CoroutineMock(return_value=returned_credentials)
            )
        )
        self.profile.context.injector.bind_instance(VCHolder, mock_vc_holder)
        record = V20PresExRecord(
            state="request-received",
            role="prover",
            pres_proposal=None,
            pres_request={
                "@type": "https://didcomm.org/present-proof/2.0/request-presentation",
                "@id": "6ae00c6c-87fa-495a-b546-5f5953817c92",
                "comment": "string",
                "formats": [
                    {
                        "attach_id": "dif",
                        "format": "dif/presentation-exchange/definitions@v1.0",
                    }
                ],
                "request_presentations~attach": [
                    {
                        "@id": "dif",
                        "mime-type": "application/json",
                        "data": {"json": test_pd},
                    }
                ],
                "will_confirm": True,
            },
            pres=None,
            verified=None,
            auto_present=False,
            error_msg=None,
        )

        with (
            mock.patch.object(
                test_module, "V20PresExRecord", autospec=True
            ) as mock_pres_ex_rec_cls,
            mock.patch.object(
                test_module.web, "json_response", mock.MagicMock()
            ) as mock_response,
        ):
            mock_pres_ex_rec_cls.retrieve_by_id.return_value = record

            await test_module.present_proof_credentials_list(self.request)
            mock_response.assert_called_once_with(
                [
                    {"name": "Credential1", "record_id": ANY},
                    {"name": "Credential2", "record_id": ANY},
                ]
            )

    async def test_present_proof_credentials_single_ldp_vp_error(self):
        self.request.match_info = {
            "pres_ex_id": "123-456-789",
        }
        self.request.query = {"extra_query": {}}
        test_pd = deepcopy(DIF_PROOF_REQ)
        test_pd["presentation_definition"]["format"] = {
            "ldp_vp": {"proof_type": ["test"]}
        }
        del test_pd["presentation_definition"]["input_descriptors"][0]["constraints"][
            "limit_disclosure"
        ]
        record = V20PresExRecord(
            state="request-received",
            role="prover",
            pres_proposal=None,
            pres_request={
                "@type": "https://didcomm.org/present-proof/2.0/request-presentation",
                "@id": "6ae00c6c-87fa-495a-b546-5f5953817c92",
                "comment": "string",
                "formats": [
                    {
                        "attach_id": "dif",
                        "format": "dif/presentation-exchange/definitions@v1.0",
                    }
                ],
                "request_presentations~attach": [
                    {
                        "@id": "dif",
                        "mime-type": "application/json",
                        "data": {"json": test_pd},
                    }
                ],
                "will_confirm": True,
            },
            pres=None,
            verified=None,
            auto_present=False,
            error_msg=None,
        )

        mock_indy_holder = mock.MagicMock(IndyHolder, autospec=True)
        mock_indy_holder.get_credentials_for_presentation_request_by_referent = (
            mock.CoroutineMock()
        )
        self.profile.context.injector.bind_instance(IndyHolder, mock_indy_holder)

        mock_vc_holder = mock.MagicMock(VCHolder, autospec=True)
        mock_vc_holder.search_credentials = mock.MagicMock()
        self.profile.context.injector.bind_instance(VCHolder, mock_vc_holder)

        with (
            mock.patch.object(
                test_module, "V20PresExRecord", autospec=True
            ) as mock_pres_ex_rec_cls,
            mock.patch.object(test_module.web, "json_response", mock.MagicMock()),
        ):
            mock_pres_ex_rec_cls.retrieve_by_id.return_value = record
            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.present_proof_credentials_list(self.request)

    async def test_present_proof_credentials_double_ldp_vp_error(self):
        self.request.match_info = {
            "pres_ex_id": "123-456-789",
        }
        self.request.query = {"extra_query": {}}
        test_pd = deepcopy(DIF_PROOF_REQ)
        test_pd["presentation_definition"]["format"] = {
            "ldp_vp": {"proof_type": ["test1", "test2"]}
        }
        del test_pd["presentation_definition"]["input_descriptors"][0]["constraints"][
            "limit_disclosure"
        ]
        record = V20PresExRecord(
            state="request-received",
            role="prover",
            pres_proposal=None,
            pres_request={
                "@type": "https://didcomm.org/present-proof/2.0/request-presentation",
                "@id": "6ae00c6c-87fa-495a-b546-5f5953817c92",
                "comment": "string",
                "formats": [
                    {
                        "attach_id": "dif",
                        "format": "dif/presentation-exchange/definitions@v1.0",
                    }
                ],
                "request_presentations~attach": [
                    {
                        "@id": "dif",
                        "mime-type": "application/json",
                        "data": {"json": test_pd},
                    }
                ],
                "will_confirm": True,
            },
            pres=None,
            verified=None,
            auto_present=False,
            error_msg=None,
        )

        mock_indy_holder = mock.MagicMock(IndyHolder, autospec=True)
        mock_indy_holder.get_credentials_for_presentation_request_by_referent = (
            mock.CoroutineMock()
        )
        self.profile.context.injector.bind_instance(IndyHolder, mock_indy_holder)

        mock_vc_holder = mock.MagicMock(VCHolder, autospec=True)
        mock_vc_holder.search_credentials = mock.MagicMock()
        self.profile.context.injector.bind_instance(VCHolder, mock_vc_holder)

        with (
            mock.patch.object(
                test_module, "V20PresExRecord", autospec=True
            ) as mock_pres_ex_rec_cls,
            mock.patch.object(test_module.web, "json_response", mock.MagicMock()),
        ):
            mock_pres_ex_rec_cls.retrieve_by_id.return_value = record
            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.present_proof_credentials_list(self.request)

    async def test_present_proof_credentials_list_limit_disclosure_no_bbs(self):
        self.request.match_info = {
            "pres_ex_id": "123-456-789",
        }
        self.request.query = {"extra_query": {}}
        test_pd = deepcopy(DIF_PROOF_REQ)
        test_pd["presentation_definition"]["format"] = {
            "ldp_vp": {"proof_type": ["Ed25519Signature2018"]}
        }
        record = V20PresExRecord(
            state="request-received",
            role="prover",
            pres_proposal=None,
            pres_request={
                "@type": "https://didcomm.org/present-proof/2.0/request-presentation",
                "@id": "6ae00c6c-87fa-495a-b546-5f5953817c92",
                "comment": "string",
                "formats": [
                    {
                        "attach_id": "dif",
                        "format": "dif/presentation-exchange/definitions@v1.0",
                    }
                ],
                "request_presentations~attach": [
                    {
                        "@id": "dif",
                        "mime-type": "application/json",
                        "data": {"json": test_pd},
                    }
                ],
                "will_confirm": True,
            },
            pres=None,
            verified=None,
            auto_present=False,
            error_msg=None,
        )

        mock_indy_holder = mock.MagicMock(IndyHolder, autospec=True)
        mock_indy_holder.get_credentials_for_presentation_request_by_referent = (
            mock.CoroutineMock()
        )
        self.profile.context.injector.bind_instance(IndyHolder, mock_indy_holder)

        mock_vc_holder = mock.MagicMock(VCHolder, autospec=True)
        mock_vc_holder.search_credentials = mock.MagicMock()
        self.profile.context.injector.bind_instance(VCHolder, mock_vc_holder)

        with (
            mock.patch.object(
                test_module, "V20PresExRecord", autospec=True
            ) as mock_pres_ex_rec_cls,
            mock.patch.object(test_module.web, "json_response", mock.MagicMock()),
        ):
            mock_pres_ex_rec_cls.retrieve_by_id.return_value = record
            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.present_proof_credentials_list(self.request)

    async def test_present_proof_credentials_no_ldp_vp(self):
        self.request.match_info = {
            "pres_ex_id": "123-456-789",
        }
        self.request.query = {"extra_query": {}}
        test_pd = deepcopy(DIF_PROOF_REQ)
        test_pd["presentation_definition"]["format"] = {
            "ldp_vc": {"proof_type": ["test"]}
        }
        del test_pd["presentation_definition"]["input_descriptors"][0]["constraints"][
            "limit_disclosure"
        ]
        record = V20PresExRecord(
            state="request-received",
            role="prover",
            pres_proposal=None,
            pres_request={
                "@type": "https://didcomm.org/present-proof/2.0/request-presentation",
                "@id": "6ae00c6c-87fa-495a-b546-5f5953817c92",
                "comment": "string",
                "formats": [
                    {
                        "attach_id": "dif",
                        "format": "dif/presentation-exchange/definitions@v1.0",
                    }
                ],
                "request_presentations~attach": [
                    {
                        "@id": "dif",
                        "mime-type": "application/json",
                        "data": {"json": test_pd},
                    }
                ],
                "will_confirm": True,
            },
            pres=None,
            verified=None,
            auto_present=False,
            error_msg=None,
        )

        mock_indy_holder = mock.MagicMock(IndyHolder, autospec=True)
        mock_indy_holder.get_credentials_for_presentation_request_by_referent = (
            mock.CoroutineMock()
        )
        self.profile.context.injector.bind_instance(IndyHolder, mock_indy_holder)

        mock_vc_holder = mock.MagicMock(VCHolder, autospec=True)
        mock_vc_holder.search_credentials = mock.MagicMock()
        self.profile.context.injector.bind_instance(VCHolder, mock_vc_holder)

        with (
            mock.patch.object(
                test_module, "V20PresExRecord", autospec=True
            ) as mock_pres_ex_rec_cls,
            mock.patch.object(test_module.web, "json_response", mock.MagicMock()),
        ):
            mock_pres_ex_rec_cls.retrieve_by_id.return_value = record
            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.present_proof_credentials_list(self.request)

    async def test_present_proof_credentials_list_schema_uri(self):
        self.request.match_info = {
            "pres_ex_id": "123-456-789",
        }
        self.request.query = {"extra_query": {}}
        test_pd = deepcopy(DIF_PROOF_REQ)
        test_pd["presentation_definition"]["input_descriptors"][0]["schema"][0]["uri"] = (
            "https://example.org/test.json"
        )
        test_pd["presentation_definition"]["input_descriptors"][0]["schema"].pop(1)
        record = V20PresExRecord(
            state="request-received",
            role="prover",
            pres_proposal=None,
            pres_request={
                "@type": "https://didcomm.org/present-proof/2.0/request-presentation",
                "@id": "6ae00c6c-87fa-495a-b546-5f5953817c92",
                "comment": "string",
                "formats": [
                    {
                        "attach_id": "dif",
                        "format": "dif/presentation-exchange/definitions@v1.0",
                    }
                ],
                "request_presentations~attach": [
                    {
                        "@id": "dif",
                        "mime-type": "application/json",
                        "data": {"json": test_pd},
                    }
                ],
                "will_confirm": True,
            },
            pres=None,
            verified=None,
            auto_present=False,
            error_msg=None,
        )

        returned_credentials = [
            mock.MagicMock(cred_value={"name": "Credential1"}),
            mock.MagicMock(cred_value={"name": "Credential2"}),
        ]
        mock_indy_holder = mock.MagicMock(IndyHolder, autospec=True)
        mock_indy_holder.get_credentials_for_presentation_request_by_referent = (
            mock.CoroutineMock()
        )
        self.profile.context.injector.bind_instance(IndyHolder, mock_indy_holder)

        mock_vc_holder = mock.MagicMock(VCHolder, autospec=True)
        mock_vc_holder.search_credentials = mock.MagicMock(
            return_value=mock.MagicMock(
                fetch=mock.CoroutineMock(return_value=returned_credentials)
            )
        )
        self.profile.context.injector.bind_instance(VCHolder, mock_vc_holder)

        with (
            mock.patch.object(
                test_module, "V20PresExRecord", autospec=True
            ) as mock_pres_ex_rec_cls,
            mock.patch.object(
                test_module.web, "json_response", mock.MagicMock()
            ) as mock_response,
        ):
            mock_pres_ex_rec_cls.retrieve_by_id.return_value = record
            await test_module.present_proof_credentials_list(self.request)
            mock_response.assert_called_once_with(
                [
                    {"name": "Credential1", "record_id": ANY},
                    {"name": "Credential2", "record_id": ANY},
                ]
            )

    async def test_present_proof_credentials_list_dif_error(self):
        self.request.match_info = {
            "pres_ex_id": "123-456-789",
        }
        self.request.query = {"extra_query": {}}

        mock_indy_holder = mock.MagicMock(IndyHolder, autospec=True)
        mock_indy_holder.get_credentials_for_presentation_request_by_referent = (
            mock.CoroutineMock()
        )
        self.profile.context.injector.bind_instance(IndyHolder, mock_indy_holder)

        mock_vc_holder = mock.MagicMock(VCHolder, autospec=True)
        mock_vc_holder.search_credentials = mock.MagicMock(
            return_value=mock.MagicMock(
                fetch=mock.CoroutineMock(side_effect=test_module.StorageNotFoundError())
            )
        )
        self.profile.context.injector.bind_instance(VCHolder, mock_vc_holder)

        record = V20PresExRecord(
            state="request-received",
            role="prover",
            pres_proposal=None,
            pres_request={
                "@type": "https://didcomm.org/present-proof/2.0/request-presentation",
                "@id": "6ae00c6c-87fa-495a-b546-5f5953817c92",
                "comment": "string",
                "formats": [
                    {
                        "attach_id": "dif",
                        "format": "dif/presentation-exchange/definitions@v1.0",
                    }
                ],
                "request_presentations~attach": [
                    {
                        "@id": "dif",
                        "mime-type": "application/json",
                        "data": {"json": DIF_PROOF_REQ},
                    }
                ],
                "will_confirm": True,
            },
            pres=None,
            verified=None,
            auto_present=False,
            error_msg=None,
        )

        with (
            mock.patch.object(
                test_module, "V20PresExRecord", autospec=True
            ) as mock_pres_ex_rec_cls,
            mock.patch.object(test_module.web, "json_response", mock.MagicMock()),
        ):
            with self.assertRaises(test_module.web.HTTPBadRequest):
                mock_pres_ex_rec_cls.retrieve_by_id.return_value = record
                await test_module.present_proof_credentials_list(self.request)

    async def test_present_proof_retrieve(self):
        self.request.match_info = {"pres_ex_id": "dummy"}

        with (
            mock.patch.object(
                test_module, "V20PresExRecord", autospec=True
            ) as mock_pres_ex_rec_cls,
            mock.patch.object(
                test_module.web, "json_response", mock.MagicMock()
            ) as mock_response,
        ):
            mock_pres_ex_rec_cls.retrieve_by_id = mock.CoroutineMock(
                return_value=mock.MagicMock(
                    serialize=mock.MagicMock(
                        return_value={"thread_id": "sample-thread-id"}
                    )
                )
            )

            await test_module.present_proof_retrieve(self.request)
            mock_response.assert_called_once_with({"thread_id": "sample-thread-id"})

    async def test_present_proof_retrieve_not_found(self):
        self.request.match_info = {"pres_ex_id": "dummy"}

        with mock.patch.object(
            test_module, "V20PresExRecord", autospec=True
        ) as mock_pres_ex_rec_cls:
            mock_pres_ex_rec_cls.retrieve_by_id = mock.CoroutineMock(
                side_effect=StorageNotFoundError()
            )

            with self.assertRaises(test_module.web.HTTPNotFound):
                await test_module.present_proof_retrieve(self.request)

    async def test_present_proof_retrieve_x(self):
        self.request.match_info = {"pres_ex_id": "dummy"}

        mock_pres_ex_rec_inst = mock.MagicMock(
            connection_id="abc123",
            thread_id="thid123",
            serialize=mock.MagicMock(side_effect=test_module.BaseModelError()),
            save_error_state=mock.CoroutineMock(),
        )
        with mock.patch.object(
            test_module, "V20PresExRecord", autospec=True
        ) as mock_pres_ex_rec_cls:
            mock_pres_ex_rec_cls.retrieve_by_id = mock.CoroutineMock(
                return_value=mock_pres_ex_rec_inst
            )

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.present_proof_retrieve(self.request)

    async def test_present_proof_send_proposal(self):
        self.request.json = mock.CoroutineMock(
            return_value={
                "connection_id": "dummy-conn-id",
                "presentation_proposal": {V20PresFormat.Format.INDY.api: INDY_PROOF_REQ},
            }
        )

        with (
            mock.patch.object(test_module, "ConnRecord", autospec=True) as mock_conn_rec,
            mock.patch.object(
                test_module, "V20PresManager", autospec=True
            ) as mock_pres_mgr,
            mock.patch.object(test_module, "V20PresExRecord", autospec=True),
            mock.patch.object(
                test_module.web, "json_response", mock.MagicMock()
            ) as mock_response,
        ):
            mock_conn_rec.retrieve_by_id = mock.CoroutineMock(
                return_value=mock.MagicMock(is_ready=True)
            )
            mock_px_rec_inst = mock.MagicMock()
            mock_pres_mgr.return_value.create_exchange_for_proposal = mock.CoroutineMock(
                return_value=mock_px_rec_inst
            )

            await test_module.present_proof_send_proposal(self.request)
            mock_response.assert_called_once_with(mock_px_rec_inst.serialize.return_value)

    async def test_present_proof_send_proposal_no_conn_record(self):
        self.request.json = mock.CoroutineMock()

        with mock.patch.object(test_module, "ConnRecord", autospec=True) as mock_conn_rec:
            mock_conn_rec.retrieve_by_id = mock.CoroutineMock(
                side_effect=StorageNotFoundError()
            )

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.present_proof_send_proposal(self.request)

    async def test_present_proof_send_proposal_not_ready(self):
        self.request.json = mock.CoroutineMock()

        with (
            mock.patch.object(
                test_module, "ConnRecord", autospec=True
            ) as mock_conn_rec_cls,
            mock.patch.object(test_module, "V20PresProposal", autospec=True),
        ):
            mock_conn_rec_cls.retrieve_by_id = mock.CoroutineMock(
                return_value=mock.MagicMock(is_ready=False)
            )

            with self.assertRaises(test_module.web.HTTPForbidden):
                await test_module.present_proof_send_proposal(self.request)

    async def test_present_proof_send_proposal_x(self):
        self.request.json = mock.CoroutineMock()

        with (
            mock.patch.object(test_module, "ConnRecord", autospec=True),
            mock.patch.object(
                test_module, "V20PresManager", autospec=True
            ) as mock_pres_mgr,
        ):
            mock_pres_mgr.return_value.create_exchange_for_proposal = mock.CoroutineMock(
                return_value=mock.MagicMock(
                    serialize=mock.MagicMock(side_effect=test_module.StorageError()),
                    save_error_state=mock.CoroutineMock(),
                )
            )

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.present_proof_send_proposal(self.request)

    async def test_present_proof_create_request(self):
        indy_proof_req = deepcopy(INDY_PROOF_REQ)
        indy_proof_req.pop("nonce")  # exercise _add_nonce()

        self.request.json = mock.CoroutineMock(
            return_value={
                "comment": "dummy",
                "presentation_request": {V20PresFormat.Format.INDY.api: indy_proof_req},
            }
        )

        with (
            mock.patch.object(
                test_module, "V20PresManager", autospec=True
            ) as mock_pres_mgr_cls,
            mock.patch.object(test_module, "V20PresRequest", autospec=True),
            mock.patch.object(
                test_module.web, "json_response", mock.MagicMock()
            ) as mock_response,
        ):
            mock_px_rec_inst = mock.MagicMock(
                serialize=mock.MagicMock(return_value={"thread_id": "sample-thread-id"})
            )
            mock_pres_mgr_inst = mock.MagicMock(
                create_exchange_for_request=mock.CoroutineMock(
                    return_value=mock_px_rec_inst
                )
            )
            mock_pres_mgr_cls.return_value = mock_pres_mgr_inst

            await test_module.present_proof_create_request(self.request)
            mock_response.assert_called_once_with(mock_px_rec_inst.serialize.return_value)

    async def test_present_proof_create_request_x(self):
        self.request.json = mock.CoroutineMock(
            return_value={
                "comment": "dummy",
                "presentation_request": {V20PresFormat.Format.INDY.api: INDY_PROOF_REQ},
            }
        )

        with (
            mock.patch.object(
                test_module, "V20PresManager", autospec=True
            ) as mock_pres_mgr_cls,
            mock.patch.object(test_module, "V20PresRequest", autospec=True),
            mock.patch.object(test_module.web, "json_response", mock.MagicMock()),
        ):
            mock_pres_mgr_inst = mock.MagicMock(
                create_exchange_for_request=mock.CoroutineMock(
                    return_value=mock.MagicMock(
                        serialize=mock.MagicMock(side_effect=test_module.StorageError()),
                        save_error_state=mock.CoroutineMock(),
                    )
                )
            )
            mock_pres_mgr_cls.return_value = mock_pres_mgr_inst

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.present_proof_create_request(self.request)

    async def test_present_proof_send_free_request(self):
        self.request.json = mock.CoroutineMock(
            return_value={
                "connection_id": "dummy",
                "comment": "dummy",
                "presentation_request": {V20PresFormat.Format.INDY.api: INDY_PROOF_REQ},
            }
        )

        with (
            mock.patch.object(
                test_module, "ConnRecord", autospec=True
            ) as mock_conn_rec_cls,
            mock.patch.object(
                test_module, "V20PresManager", autospec=True
            ) as mock_pres_mgr_cls,
            mock.patch.object(test_module, "V20PresRequest", autospec=True),
            mock.patch.object(test_module, "V20PresExRecord", autospec=True),
            mock.patch.object(
                test_module.web, "json_response", mock.MagicMock()
            ) as mock_response,
        ):
            mock_conn_rec_cls.retrieve_by_id = mock.CoroutineMock()
            mock_px_rec_inst = mock.MagicMock(
                serialize=mock.MagicMock({"thread_id": "sample-thread-id"})
            )

            mock_pres_mgr_inst = mock.MagicMock(
                create_exchange_for_request=mock.CoroutineMock(
                    return_value=mock_px_rec_inst
                )
            )
            mock_pres_mgr_cls.return_value = mock_pres_mgr_inst

            await test_module.present_proof_send_free_request(self.request)
            mock_response.assert_called_once_with(mock_px_rec_inst.serialize.return_value)

    async def test_present_proof_send_free_request_not_found(self):
        self.request.json = mock.CoroutineMock(return_value={"connection_id": "dummy"})

        with mock.patch.object(
            test_module, "ConnRecord", mock.MagicMock()
        ) as mock_conn_rec_cls:
            mock_conn_rec_cls.retrieve_by_id = mock.CoroutineMock(
                side_effect=StorageNotFoundError()
            )

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.present_proof_send_free_request(self.request)

    async def test_present_proof_send_free_request_not_ready(self):
        self.request.json = mock.CoroutineMock(
            return_value={"connection_id": "dummy", "proof_request": {}}
        )

        with mock.patch.object(
            test_module, "ConnRecord", mock.MagicMock()
        ) as mock_conn_rec_cls:
            mock_conn_rec_inst = mock.MagicMock(is_ready=False)
            mock_conn_rec_cls.retrieve_by_id = mock.CoroutineMock(
                return_value=mock_conn_rec_inst
            )

            with self.assertRaises(test_module.web.HTTPForbidden):
                await test_module.present_proof_send_free_request(self.request)

    async def test_present_proof_send_free_request_x(self):
        self.request.json = mock.CoroutineMock(
            return_value={
                "connection_id": "dummy",
                "comment": "dummy",
                "presentation_request": {V20PresFormat.Format.INDY.api: INDY_PROOF_REQ},
            }
        )

        with (
            mock.patch.object(
                test_module, "ConnRecord", autospec=True
            ) as mock_conn_rec_cls,
            mock.patch.object(
                test_module, "V20PresManager", autospec=True
            ) as mock_pres_mgr_cls,
            mock.patch.object(test_module, "V20PresRequest", autospec=True),
            mock.patch.object(test_module, "V20PresExRecord", autospec=True),
        ):
            mock_conn_rec_inst = mock.MagicMock()
            mock_conn_rec_cls.retrieve_by_id = mock.CoroutineMock(
                return_value=mock_conn_rec_inst
            )
            mock_pres_mgr_inst = mock.MagicMock(
                create_exchange_for_request=mock.CoroutineMock(
                    return_value=mock.MagicMock(
                        serialize=mock.MagicMock(side_effect=test_module.StorageError()),
                        save_error_state=mock.CoroutineMock(),
                    )
                )
            )
            mock_pres_mgr_cls.return_value = mock_pres_mgr_inst

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.present_proof_send_free_request(self.request)

    async def test_present_proof_send_bound_request(self):
        self.request.json = mock.CoroutineMock(return_value={"trace": False})
        self.request.match_info = {"pres_ex_id": "dummy"}

        self.profile.context.injector.bind_instance(
            BaseLedger,
            mock.MagicMock(
                __aenter__=mock.CoroutineMock(),
                __aexit__=mock.CoroutineMock(),
            ),
        )
        self.profile.context.injector.bind_instance(
            IndyVerifier,
            mock.MagicMock(
                verify_presentation=mock.CoroutineMock(),
            ),
        )

        with (
            mock.patch.object(
                test_module, "ConnRecord", autospec=True
            ) as mock_conn_rec_cls,
            mock.patch.object(
                test_module, "V20PresManager", autospec=True
            ) as mock_pres_mgr_cls,
            mock.patch.object(
                test_module, "V20PresExRecord", autospec=True
            ) as mock_px_rec_cls,
            mock.patch.object(
                test_module.web, "json_response", mock.MagicMock()
            ) as mock_response,
        ):
            mock_px_rec_inst = mock.MagicMock(
                connection_id="dummy",
                state=test_module.V20PresExRecord.STATE_PROPOSAL_RECEIVED,
                serialize=mock.MagicMock(return_value={"thread_id": "sample-thread-id"}),
            )
            mock_px_rec_cls.retrieve_by_id = mock.CoroutineMock(
                return_value=mock_px_rec_inst
            )
            mock_conn_rec_inst = mock.MagicMock(
                is_ready=True,
            )
            mock_conn_rec_cls.retrieve_by_id = mock.CoroutineMock(
                return_value=mock_conn_rec_inst
            )
            mock_pres_request = mock.MagicMock()

            mock_pres_mgr_inst = mock.MagicMock(
                create_bound_request=mock.CoroutineMock(
                    return_value=(mock_px_rec_inst, mock_pres_request)
                )
            )
            mock_pres_mgr_cls.return_value = mock_pres_mgr_inst

            await test_module.present_proof_send_bound_request(self.request)
            mock_response.assert_called_once_with(mock_px_rec_inst.serialize.return_value)

    async def test_present_proof_send_bound_request_not_found(self):
        self.request.json = mock.CoroutineMock(return_value={"trace": False})
        self.request.match_info = {"pres_ex_id": "dummy"}

        self.profile.context.injector.bind_instance(
            BaseLedger,
            mock.MagicMock(
                __aenter__=mock.CoroutineMock(),
                __aexit__=mock.CoroutineMock(),
            ),
        )
        self.profile.context.injector.bind_instance(
            IndyVerifier,
            mock.MagicMock(
                verify_presentation=mock.CoroutineMock(),
            ),
        )

        with (
            mock.patch.object(
                test_module, "ConnRecord", autospec=True
            ) as mock_conn_rec_cls,
            mock.patch.object(
                test_module, "V20PresExRecord", autospec=True
            ) as mock_px_rec_cls,
        ):
            mock_px_rec_inst = mock.MagicMock(
                connection_id="dummy",
                state=test_module.V20PresExRecord.STATE_PROPOSAL_RECEIVED,
                serialize=mock.MagicMock(return_value={"thread_id": "sample-thread-id"}),
            )
            mock_px_rec_cls.retrieve_by_id = mock.CoroutineMock(
                return_value=mock_px_rec_inst
            )
            mock_conn_rec_cls.retrieve_by_id = mock.CoroutineMock(
                side_effect=StorageNotFoundError()
            )

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.present_proof_send_bound_request(self.request)

    async def test_present_proof_send_bound_request_not_ready(self):
        self.request.json = mock.CoroutineMock(return_value={"trace": False})
        self.request.match_info = {"pres_ex_id": "dummy"}

        self.profile.context.injector.bind_instance(
            BaseLedger,
            mock.MagicMock(
                __aenter__=mock.CoroutineMock(),
                __aexit__=mock.CoroutineMock(),
            ),
        )
        self.profile.context.injector.bind_instance(
            IndyVerifier,
            mock.MagicMock(
                verify_presentation=mock.CoroutineMock(),
            ),
        )

        with (
            mock.patch.object(
                test_module, "ConnRecord", autospec=True
            ) as mock_conn_rec_cls,
            mock.patch.object(
                test_module, "V20PresExRecord", autospec=True
            ) as mock_px_rec_cls,
        ):
            mock_px_rec_inst = mock.MagicMock(
                connection_id="dummy",
                state=test_module.V20PresExRecord.STATE_PROPOSAL_RECEIVED,
                serialize=mock.MagicMock(return_value={"thread_id": "sample-thread-id"}),
            )
            mock_px_rec_cls.retrieve_by_id = mock.CoroutineMock(
                return_value=mock_px_rec_inst
            )
            mock_conn_rec_inst = mock.MagicMock(
                is_ready=False,
            )
            mock_conn_rec_cls.retrieve_by_id = mock.CoroutineMock(
                return_value=mock_conn_rec_inst
            )

            with self.assertRaises(test_module.web.HTTPForbidden):
                await test_module.present_proof_send_bound_request(self.request)

    async def test_present_proof_send_bound_request_px_rec_not_found(self):
        self.request.json = mock.CoroutineMock(return_value={"trace": False})
        self.request.match_info = {"pres_ex_id": "dummy"}

        with mock.patch.object(
            test_module, "V20PresExRecord", autospec=True
        ) as mock_px_rec_cls:
            mock_px_rec_cls.retrieve_by_id = mock.CoroutineMock(
                side_effect=StorageNotFoundError("no such record")
            )
            with self.assertRaises(test_module.web.HTTPNotFound) as context:
                await test_module.present_proof_send_bound_request(self.request)
            assert "no such record" in str(context.exception)

    async def test_present_proof_send_bound_request_bad_state(self):
        self.request.json = mock.CoroutineMock(return_value={"trace": False})
        self.request.match_info = {"pres_ex_id": "dummy"}

        self.profile.context.injector.bind_instance(
            BaseLedger,
            mock.MagicMock(
                __aenter__=mock.CoroutineMock(),
                __aexit__=mock.CoroutineMock(),
            ),
        )
        self.profile.context.injector.bind_instance(
            IndyVerifier,
            mock.MagicMock(
                verify_presentation=mock.CoroutineMock(),
            ),
        )

        with mock.patch.object(
            test_module, "V20PresExRecord", autospec=True
        ) as mock_px_rec_cls:
            mock_px_rec_inst = mock.MagicMock(
                connection_id="dummy",
                state=test_module.V20PresExRecord.STATE_DONE,
                serialize=mock.MagicMock(return_value={"thread_id": "sample-thread-id"}),
            )
            mock_px_rec_cls.retrieve_by_id = mock.CoroutineMock(
                return_value=mock_px_rec_inst
            )

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.present_proof_send_bound_request(self.request)

    async def test_present_proof_send_bound_request_x(self):
        self.request.json = mock.CoroutineMock(return_value={"trace": False})
        self.request.match_info = {"pres_ex_id": "dummy"}

        self.profile.context.injector.bind_instance(
            BaseLedger,
            mock.MagicMock(
                __aenter__=mock.CoroutineMock(),
                __aexit__=mock.CoroutineMock(),
            ),
        )
        self.profile.context.injector.bind_instance(
            IndyVerifier,
            mock.MagicMock(
                verify_presentation=mock.CoroutineMock(),
            ),
        )

        with (
            mock.patch.object(
                test_module, "ConnRecord", autospec=True
            ) as mock_conn_rec_cls,
            mock.patch.object(
                test_module, "V20PresManager", autospec=True
            ) as mock_pres_mgr_cls,
            mock.patch.object(
                test_module, "V20PresExRecord", autospec=True
            ) as mock_px_rec_cls,
        ):
            mock_px_rec_inst = mock.MagicMock(
                connection_id="dummy",
                state=test_module.V20PresExRecord.STATE_PROPOSAL_RECEIVED,
                serialize=mock.MagicMock(return_value={"thread_id": "sample-thread-id"}),
                save_error_state=mock.CoroutineMock(),
            )
            mock_px_rec_cls.retrieve_by_id = mock.CoroutineMock(
                return_value=mock_px_rec_inst
            )
            mock_conn_rec_inst = mock.MagicMock(
                is_ready=True,
            )
            mock_conn_rec_cls.retrieve_by_id = mock.CoroutineMock(
                return_value=mock_conn_rec_inst
            )

            mock_pres_mgr_inst = mock.MagicMock(
                create_bound_request=mock.CoroutineMock(
                    side_effect=test_module.StorageError()
                )
            )
            mock_pres_mgr_cls.return_value = mock_pres_mgr_inst

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.present_proof_send_bound_request(self.request)

    async def test_present_proof_send_presentation(self):
        self.request.json = mock.CoroutineMock(
            return_value={
                "indy": {
                    "comment": "dummy",
                    "self_attested_attributes": {},
                    "requested_attributes": {},
                    "requested_predicates": {},
                }
            }
        )
        self.request.match_info = {
            "pres_ex_id": "dummy",
        }
        self.profile.context.injector.bind_instance(
            IndyVerifier,
            mock.MagicMock(
                verify_presentation=mock.CoroutineMock(),
            ),
        )

        with (
            mock.patch.object(
                test_module, "ConnRecord", autospec=True
            ) as mock_conn_rec_cls,
            mock.patch.object(
                test_module, "V20PresManager", autospec=True
            ) as mock_pres_mgr_cls,
            mock.patch.object(
                test_module, "V20PresExRecord", autospec=True
            ) as mock_px_rec_cls,
            mock.patch.object(test_module.web, "json_response") as mock_response,
        ):
            mock_px_rec_inst = mock.MagicMock(
                connection_id="dummy",
                state=test_module.V20PresExRecord.STATE_REQUEST_RECEIVED,
                serialize=mock.MagicMock(return_value={"thread_id": "sample-thread-id"}),
            )
            mock_px_rec_cls.retrieve_by_id = mock.CoroutineMock(
                return_value=mock_px_rec_inst
            )

            mock_conn_rec_inst = mock.MagicMock(is_ready=True)
            mock_conn_rec_cls.retrieve_by_id = mock.CoroutineMock(
                return_value=mock_conn_rec_inst
            )

            mock_pres_mgr_inst = mock.MagicMock(
                create_pres=mock.CoroutineMock(
                    return_value=(mock_px_rec_inst, mock.MagicMock())
                )
            )
            mock_pres_mgr_cls.return_value = mock_pres_mgr_inst

            await test_module.present_proof_send_presentation(self.request)
            mock_response.assert_called_once_with(mock_px_rec_inst.serialize.return_value)

    async def test_present_proof_send_presentation_dif(self):
        proof_req = deepcopy(DIF_PROOF_REQ)
        proof_req["issuer_id"] = "test123"
        self.request.json = mock.CoroutineMock(
            return_value={
                "dif": proof_req,
            }
        )
        self.request.match_info = {
            "pres_ex_id": "dummy",
        }
        self.profile.context.injector.bind_instance(
            IndyVerifier,
            mock.MagicMock(
                verify_presentation=mock.CoroutineMock(),
            ),
        )

        with (
            mock.patch.object(
                test_module, "ConnRecord", autospec=True
            ) as mock_conn_rec_cls,
            mock.patch.object(
                test_module, "V20PresManager", autospec=True
            ) as mock_pres_mgr_cls,
            mock.patch.object(
                test_module, "V20PresExRecord", autospec=True
            ) as mock_px_rec_cls,
            mock.patch.object(test_module.web, "json_response") as mock_response,
        ):
            mock_px_rec_inst = mock.MagicMock(
                connection_id="dummy",
                state=test_module.V20PresExRecord.STATE_REQUEST_RECEIVED,
                serialize=mock.MagicMock(return_value={"thread_id": "sample-thread-id"}),
            )
            mock_px_rec_cls.retrieve_by_id = mock.CoroutineMock(
                return_value=mock_px_rec_inst
            )

            mock_conn_rec_inst = mock.MagicMock(is_ready=True)
            mock_conn_rec_cls.retrieve_by_id = mock.CoroutineMock(
                return_value=mock_conn_rec_inst
            )

            mock_pres_mgr_inst = mock.MagicMock(
                create_pres=mock.CoroutineMock(
                    return_value=(mock_px_rec_inst, mock.MagicMock())
                )
            )
            mock_pres_mgr_cls.return_value = mock_pres_mgr_inst

            await test_module.present_proof_send_presentation(self.request)
            mock_response.assert_called_once_with(mock_px_rec_inst.serialize.return_value)

    async def test_present_proof_send_presentation_dif_error(self):
        self.request.json = mock.CoroutineMock(return_value={"dif": DIF_PROOF_REQ})
        self.request.match_info = {
            "pres_ex_id": "dummy",
        }
        px_rec_instance = V20PresExRecord(
            state="request-received",
            role="prover",
            pres_proposal=None,
            pres_request={
                "@type": "https://didcomm.org/present-proof/2.0/request-presentation",
                "@id": "6ae00c6c-87fa-495a-b546-5f5953817c92",
                "comment": "string",
                "formats": [
                    {
                        "attach_id": "dif",
                        "format": "dif/presentation-exchange/definitions@v1.0",
                    }
                ],
                "request_presentations~attach": [
                    {
                        "@id": "dif",
                        "mime-type": "application/json",
                        "data": {"json": DIF_PROOF_REQ},
                    }
                ],
                "will_confirm": True,
            },
            pres=None,
            verified=None,
            auto_present=False,
            error_msg=None,
        )
        self.profile.context.injector.bind_instance(
            IndyVerifier,
            mock.MagicMock(
                verify_presentation=mock.CoroutineMock(),
            ),
        )

        with (
            mock.patch.object(
                test_module, "ConnRecord", autospec=True
            ) as mock_conn_rec_cls,
            mock.patch.object(
                test_module, "V20PresManager", autospec=True
            ) as mock_pres_mgr_cls,
            mock.patch.object(
                test_module, "V20PresExRecord", autospec=True
            ) as mock_px_rec_cls,
            mock.patch.object(test_module.web, "json_response") as mock_response,
        ):
            mock_px_rec_cls.retrieve_by_id = mock.CoroutineMock(
                return_value=px_rec_instance
            )

            mock_conn_rec_inst = mock.MagicMock(is_ready=True)
            mock_conn_rec_cls.retrieve_by_id = mock.CoroutineMock(
                return_value=mock_conn_rec_inst
            )

            mock_pres_mgr_inst = mock.MagicMock(
                create_pres=mock.CoroutineMock(side_effect=test_module.LedgerError())
            )
            mock_pres_mgr_cls.return_value = mock_pres_mgr_inst
            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.present_proof_send_presentation(self.request)
                mock_response.assert_called_once_with(px_rec_instance.serialize())

    async def test_present_proof_send_presentation_px_rec_not_found(self):
        self.request.json = mock.CoroutineMock(
            return_value={
                "indy": {
                    "comment": "dummy",
                    "self_attested_attributes": {},
                    "requested_attributes": {},
                    "requested_predicates": {},
                }
            }
        )
        self.request.match_info = {
            "pres_ex_id": "dummy",
        }

        with mock.patch.object(
            test_module, "V20PresExRecord", autospec=True
        ) as mock_px_rec_cls:
            mock_px_rec_cls.retrieve_by_id = mock.CoroutineMock(
                side_effect=StorageNotFoundError("no such record")
            )

            with self.assertRaises(test_module.web.HTTPNotFound) as context:
                await test_module.present_proof_send_presentation(self.request)
            assert "no such record" in str(context.exception)

    async def test_present_proof_send_presentation_not_found(self):
        self.request.json = mock.CoroutineMock(
            return_value={
                "indy": {
                    "comment": "dummy",
                    "self_attested_attributes": {},
                    "requested_attributes": {},
                    "requested_predicates": {},
                }
            }
        )
        self.request.match_info = {
            "pres_ex_id": "dummy",
        }
        self.profile.context.injector.bind_instance(
            IndyVerifier,
            mock.MagicMock(
                verify_presentation=mock.CoroutineMock(),
            ),
        )

        with (
            mock.patch.object(
                test_module, "ConnRecord", autospec=True
            ) as mock_conn_rec_cls,
            mock.patch.object(
                test_module, "V20PresExRecord", autospec=True
            ) as mock_px_rec_cls,
        ):
            mock_px_rec_inst = mock.MagicMock(
                connection_id="dummy",
                state=test_module.V20PresExRecord.STATE_REQUEST_RECEIVED,
                serialize=mock.MagicMock(return_value={"thread_id": "sample-thread-id"}),
            )
            mock_px_rec_cls.retrieve_by_id = mock.CoroutineMock(
                return_value=mock_px_rec_inst
            )

            mock_conn_rec_cls.retrieve_by_id = mock.CoroutineMock(
                side_effect=StorageNotFoundError()
            )

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.present_proof_send_presentation(self.request)

    async def test_present_proof_send_presentation_not_ready(self):
        self.request.json = mock.CoroutineMock(
            return_value={
                "indy": {
                    "comment": "dummy",
                    "self_attested_attributes": {},
                    "requested_attributes": {},
                    "requested_predicates": {},
                }
            }
        )
        self.request.match_info = {
            "pres_ex_id": "dummy",
        }
        self.profile.context.injector.bind_instance(
            IndyVerifier,
            mock.MagicMock(
                verify_presentation=mock.CoroutineMock(),
            ),
        )

        with (
            mock.patch.object(
                test_module, "ConnRecord", autospec=True
            ) as mock_conn_rec_cls,
            mock.patch.object(
                test_module, "V20PresExRecord", autospec=True
            ) as mock_px_rec_cls,
        ):
            mock_px_rec_inst = mock.MagicMock(
                connection_id="dummy",
                state=test_module.V20PresExRecord.STATE_REQUEST_RECEIVED,
                serialize=mock.MagicMock(return_value={"thread_id": "sample-thread-id"}),
            )
            mock_px_rec_cls.retrieve_by_id = mock.CoroutineMock(
                return_value=mock_px_rec_inst
            )

            mock_conn_rec_cls.retrieve_by_id = mock.CoroutineMock(
                return_value=mock.MagicMock(is_ready=False)
            )

            with self.assertRaises(test_module.web.HTTPForbidden):
                await test_module.present_proof_send_presentation(self.request)

    async def test_present_proof_send_presentation_bad_state(self):
        self.request.json = mock.CoroutineMock(
            return_value={
                "indy": {
                    "comment": "dummy",
                    "self_attested_attributes": {},
                    "requested_attributes": {},
                    "requested_predicates": {},
                }
            }
        )
        self.request.match_info = {
            "pres_ex_id": "dummy",
        }

        with mock.patch.object(
            test_module, "V20PresExRecord", autospec=True
        ) as mock_px_rec_cls:
            mock_px_rec_inst = mock.MagicMock(
                connection_id=None,
                state=test_module.V20PresExRecord.STATE_DONE,
                serialize=mock.MagicMock(return_value={"thread_id": "sample-thread-id"}),
            )
            mock_px_rec_cls.retrieve_by_id = mock.CoroutineMock(
                return_value=mock_px_rec_inst
            )

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.present_proof_send_presentation(self.request)

    async def test_present_proof_send_presentation_x(self):
        self.request.json = mock.CoroutineMock(
            return_value={
                "indy": {
                    "comment": "dummy",
                    "self_attested_attributes": {},
                    "requested_attributes": {},
                    "requested_predicates": {},
                }
            }
        )
        self.request.match_info = {
            "pres_ex_id": "dummy",
        }
        self.profile.context.injector.bind_instance(
            IndyVerifier,
            mock.MagicMock(
                verify_presentation=mock.CoroutineMock(),
            ),
        )

        with (
            mock.patch.object(
                test_module, "ConnRecord", autospec=True
            ) as mock_conn_rec_cls,
            mock.patch.object(
                test_module, "V20PresManager", autospec=True
            ) as mock_pres_mgr_cls,
            mock.patch.object(
                test_module, "V20PresExRecord", autospec=True
            ) as mock_px_rec_cls,
            mock.patch.object(test_module.web, "json_response"),
        ):
            mock_px_rec_inst = mock.MagicMock(
                connection_id="dummy",
                state=test_module.V20PresExRecord.STATE_REQUEST_RECEIVED,
                serialize=mock.MagicMock(return_value={"thread_id": "sample-thread-id"}),
                save_error_state=mock.CoroutineMock(),
            )
            mock_px_rec_cls.retrieve_by_id = mock.CoroutineMock(
                return_value=mock_px_rec_inst
            )

            mock_conn_rec_inst = mock.MagicMock(is_ready=True)
            mock_conn_rec_cls.retrieve_by_id = mock.CoroutineMock(
                return_value=mock_conn_rec_inst
            )

            mock_pres_mgr_inst = mock.MagicMock(
                create_pres=mock.CoroutineMock(
                    side_effect=[
                        test_module.LedgerError(),
                        test_module.StorageError(),
                    ]
                )
            )
            mock_pres_mgr_cls.return_value = mock_pres_mgr_inst

            with self.assertRaises(test_module.web.HTTPBadRequest):  # ledger error
                await test_module.present_proof_send_presentation(self.request)
            with self.assertRaises(test_module.web.HTTPBadRequest):  # storage error
                await test_module.present_proof_send_presentation(self.request)

    async def test_present_proof_verify_presentation(self):
        self.request.match_info = {"pres_ex_id": "dummy"}

        with (
            mock.patch.object(
                test_module, "ConnRecord", autospec=True
            ) as mock_conn_rec_cls,
            mock.patch.object(
                test_module, "V20PresManager", autospec=True
            ) as mock_pres_mgr_cls,
            mock.patch.object(
                test_module, "V20PresExRecord", autospec=True
            ) as mock_px_rec_cls,
            mock.patch.object(
                test_module.web, "json_response", mock.MagicMock()
            ) as mock_response,
        ):
            mock_px_rec_inst = mock.MagicMock(
                connection_id="dummy",
                state=test_module.V20PresExRecord.STATE_PRESENTATION_RECEIVED,
                serialize=mock.MagicMock(return_value={"thread_id": "sample-thread-id"}),
            )
            mock_px_rec_cls.retrieve_by_id = mock.CoroutineMock(
                return_value=mock_px_rec_inst
            )
            mock_conn_rec_inst = mock.MagicMock(is_ready=True)
            mock_conn_rec_cls.retrieve_by_id = mock.CoroutineMock(
                return_value=mock_conn_rec_inst
            )

            mock_pres_mgr_inst = mock.MagicMock(
                verify_pres=mock.CoroutineMock(return_value=mock_px_rec_inst)
            )
            mock_pres_mgr_cls.return_value = mock_pres_mgr_inst

            await test_module.present_proof_verify_presentation(self.request)
            mock_response.assert_called_once_with({"thread_id": "sample-thread-id"})

    async def test_present_proof_verify_presentation_px_rec_not_found(self):
        self.request.match_info = {"pres_ex_id": "dummy"}

        with mock.patch.object(
            test_module, "V20PresExRecord", autospec=True
        ) as mock_px_rec_cls:
            mock_px_rec_cls.retrieve_by_id = mock.CoroutineMock(
                side_effect=StorageNotFoundError("no such record")
            )

            with self.assertRaises(test_module.web.HTTPNotFound) as context:
                await test_module.present_proof_verify_presentation(self.request)
            assert "no such record" in str(context.exception)

    async def test_present_proof_verify_presentation_bad_state(self):
        self.request.match_info = {"pres_ex_id": "dummy"}

        with mock.patch.object(
            test_module, "V20PresExRecord", autospec=True
        ) as mock_px_rec_cls:
            mock_px_rec_inst = mock.MagicMock(
                connection_id="dummy",
                state=test_module.V20PresExRecord.STATE_DONE,
                serialize=mock.MagicMock(return_value={"thread_id": "sample-thread-id"}),
            )
            mock_px_rec_cls.retrieve_by_id = mock.CoroutineMock(
                return_value=mock_px_rec_inst
            )

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.present_proof_verify_presentation(self.request)

    async def test_present_proof_verify_presentation_x(self):
        self.request.match_info = {"pres_ex_id": "dummy"}

        with (
            mock.patch.object(
                test_module, "ConnRecord", autospec=True
            ) as mock_conn_rec_cls,
            mock.patch.object(
                test_module, "V20PresManager", autospec=True
            ) as mock_pres_mgr_cls,
            mock.patch.object(
                test_module, "V20PresExRecord", autospec=True
            ) as mock_px_rec_cls,
            mock.patch.object(test_module.web, "json_response", mock.MagicMock()),
        ):
            mock_px_rec_inst = mock.MagicMock(
                connection_id="dummy",
                state=test_module.V20PresExRecord.STATE_PRESENTATION_RECEIVED,
                serialize=mock.MagicMock(return_value={"thread_id": "sample-thread-id"}),
                save_error_state=mock.CoroutineMock(),
            )
            mock_px_rec_cls.retrieve_by_id = mock.CoroutineMock(
                return_value=mock_px_rec_inst
            )
            mock_conn_rec_inst = mock.MagicMock(is_ready=True)
            mock_conn_rec_cls.retrieve_by_id = mock.CoroutineMock(
                return_value=mock_conn_rec_inst
            )

            mock_pres_mgr_inst = mock.MagicMock(
                verify_pres=mock.CoroutineMock(
                    side_effect=[
                        test_module.LedgerError(),
                        test_module.StorageError(),
                    ]
                )
            )
            mock_pres_mgr_cls.return_value = mock_pres_mgr_inst

            with self.assertRaises(test_module.web.HTTPBadRequest):  # ledger error
                await test_module.present_proof_verify_presentation(self.request)
            with self.assertRaises(test_module.web.HTTPBadRequest):  # storage error
                await test_module.present_proof_verify_presentation(self.request)

    async def test_present_proof_problem_report(self):
        self.request.json = mock.CoroutineMock(
            return_value={"description": "Did I say no problem? I meant 'No! Problem.'"}
        )
        self.request.match_info = {"pres_ex_id": "dummy"}
        magic_report = mock.MagicMock()

        with (
            mock.patch.object(test_module, "V20PresManager", autospec=True),
            mock.patch.object(
                test_module, "problem_report_for_record", mock.MagicMock()
            ) as mock_problem_report,
            mock.patch.object(
                test_module, "V20PresExRecord", autospec=True
            ) as mock_px_rec,
            mock.patch.object(test_module.web, "json_response") as mock_response,
        ):
            mock_px_rec.retrieve_by_id = mock.CoroutineMock(
                return_value=mock.MagicMock(save_error_state=mock.CoroutineMock())
            )
            mock_problem_report.return_value = magic_report

            await test_module.present_proof_problem_report(self.request)

            self.request["outbound_message_router"].assert_awaited_once()
            mock_response.assert_called_once_with({})

    async def test_present_proof_problem_report_bad_pres_ex_id(self):
        self.request.json = mock.CoroutineMock(
            return_value={"description": "Did I say no problem? I meant 'No! Problem.'"}
        )
        self.request.match_info = {"pres_ex_id": "dummy"}

        with mock.patch.object(
            test_module, "V20PresExRecord", autospec=True
        ) as mock_px_rec:
            mock_px_rec.retrieve_by_id = mock.CoroutineMock(
                side_effect=test_module.StorageNotFoundError()
            )

            with self.assertRaises(test_module.web.HTTPNotFound):
                await test_module.present_proof_problem_report(self.request)

    async def test_present_proof_problem_report_x(self):
        self.request.json = mock.CoroutineMock(
            return_value={"description": "Did I say no problem? I meant 'No! Problem.'"}
        )
        self.request.match_info = {"pres_ex_id": "dummy"}

        with (
            mock.patch.object(test_module, "V20PresManager", autospec=True),
            mock.patch.object(test_module, "problem_report_for_record", mock.MagicMock()),
            mock.patch.object(
                test_module, "V20PresExRecord", autospec=True
            ) as mock_px_rec,
        ):
            mock_px_rec.retrieve_by_id = mock.CoroutineMock(
                side_effect=test_module.StorageError()
            )

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.present_proof_problem_report(self.request)

    async def test_present_proof_remove(self):
        self.request.match_info = {"pres_ex_id": "dummy"}

        with (
            mock.patch.object(
                test_module, "V20PresExRecord", autospec=True
            ) as mock_px_rec,
            mock.patch.object(
                test_module.web, "json_response", mock.MagicMock()
            ) as mock_response,
        ):
            mock_px_rec.retrieve_by_id = mock.CoroutineMock(
                return_value=mock.MagicMock(
                    state=test_module.V20PresExRecord.STATE_DONE,
                    connection_id="dummy",
                    delete_record=mock.CoroutineMock(),
                )
            )

            await test_module.present_proof_remove(self.request)
            mock_response.assert_called_once_with({})

    async def test_present_proof_remove_px_rec_not_found(self):
        self.request.match_info = {"pres_ex_id": "dummy"}

        with mock.patch.object(
            test_module, "V20PresExRecord", autospec=True
        ) as mock_px_rec:
            mock_px_rec.retrieve_by_id = mock.CoroutineMock(
                side_effect=StorageNotFoundError()
            )

            with self.assertRaises(test_module.web.HTTPNotFound):
                await test_module.present_proof_remove(self.request)

    async def test_present_proof_remove_x(self):
        self.request.match_info = {"pres_ex_id": "dummy"}

        with mock.patch.object(
            test_module, "V20PresExRecord", autospec=True
        ) as mock_px_rec:
            mock_px_rec.retrieve_by_id = mock.CoroutineMock(
                return_value=mock.MagicMock(
                    state=test_module.V20PresExRecord.STATE_DONE,
                    connection_id="dummy",
                    delete_record=mock.CoroutineMock(
                        side_effect=test_module.StorageError()
                    ),
                )
            )

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.present_proof_remove(self.request)

    async def test_register(self):
        mock_app = mock.MagicMock()
        mock_app.add_routes = mock.MagicMock()

        await test_module.register(mock_app)
        mock_app.add_routes.assert_called_once()

    async def test_post_process_routes(self):
        mock_app = mock.MagicMock(_state={"swagger_dict": {}})
        test_module.post_process_routes(mock_app)
        assert "tags" in mock_app._state["swagger_dict"]

    def test_format_attach_dif(self):
        req_dict = {"dif": DIF_PROOF_REQ}
        pres_req_dict = test_module._formats_attach(
            by_format=req_dict,
            msg_type="present-proof/2.0/request-presentation",
            spec="request_presentations",
        )
        assert pres_req_dict.get("formats")[0].attach_id == "dif"
        assert (
            pres_req_dict.get("request_presentations_attach")[0].data.json_
            == DIF_PROOF_REQ
        )

    async def test_process_vcrecords_return_list(self):
        cred_list = [
            VCRecord(
                contexts=[
                    "https://www.w3.org/2018/credentials/v1",
                    "https://www.w3.org/2018/credentials/examples/v1",
                ],
                expanded_types=[
                    "https://www.w3.org/2018/credentials#VerifiableCredential",
                    "https://example.org/examples#UniversityDegreeCredential",
                ],
                issuer_id="https://example.edu/issuers/565049",
                subject_ids=[
                    "did:sov:LjgpST2rjsoxYegQDRm7EL",
                    "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL",
                ],
                proof_types=["BbsBlsSignature2020"],
                schema_ids=["https://example.org/examples/degree.json"],
                cred_value={"...": "..."},
                given_id="http://example.edu/credentials/3732",
                cred_tags={"some": "tag"},
                record_id="test1",
            ),
            VCRecord(
                contexts=[
                    "https://www.w3.org/2018/credentials/v1",
                    "https://www.w3.org/2018/credentials/examples/v1",
                ],
                expanded_types=[
                    "https://www.w3.org/2018/credentials#VerifiableCredential",
                    "https://example.org/examples#UniversityDegreeCredential",
                ],
                issuer_id="https://example.edu/issuers/565049",
                subject_ids=[
                    "did:sov:LjgpST2rjsoxYegQDRm7EL",
                    "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL",
                ],
                proof_types=["BbsBlsSignature2020"],
                schema_ids=["https://example.org/examples/degree.json"],
                cred_value={"...": "..."},
                given_id="http://example.edu/credentials/3732",
                cred_tags={"some": "tag"},
                record_id="test2",
            ),
        ]
        record_ids = {"test1"}
        (
            returned_cred_list,
            returned_record_ids,
        ) = await test_module.process_vcrecords_return_list(cred_list, record_ids)
        assert len(returned_cred_list) == 1
        assert len(returned_record_ids) == 2
        assert returned_cred_list[0].record_id == "test2"

    async def test_retrieve_uri_list_from_schema_filter(self):
        test_schema_filter = [
            [
                SchemaInputDescriptor(uri="test123"),
                SchemaInputDescriptor(uri="test321", required=True),
            ]
        ]
        test_one_of_uri_groups = await test_module.retrieve_uri_list_from_schema_filter(
            test_schema_filter
        )
        assert test_one_of_uri_groups == [["test123", "test321"]]

    async def test_send_presentation_no_specification(self):
        self.request.json = mock.CoroutineMock(return_value={"comment": "test"})
        self.request.match_info = {
            "pres_ex_id": "dummy",
        }
        with self.assertRaises(test_module.web.HTTPBadRequest):
            await test_module.present_proof_send_presentation(self.request)

    async def test_v20presentationsendreqschema(self):
        test_input = {
            "comment": "string",
            "connection_id": "631522e9-ca17-4c88-9a4c-d1cad35e463a",
            "presentation_request": {
                "dif": {
                    "_schema": [
                        {
                            "uri": "https://www.w3.org/2018/credentials/#VerifiableCredential"
                        }
                    ],
                    "presentation_definition": {
                        "format": {"ldp_vp": {"proof_type": "BbsBlsSignature2020"}},
                        "id": "fa2c4a76-c7bd-4313-a0f8-d9f5979c1fd2",
                        "input_descriptors": [
                            {
                                "schema": [
                                    {
                                        "uri": "https://www.w3.org/2018/credentials/#VerifiableCredential"
                                    }
                                ],
                                "constraints": {
                                    "fields": [
                                        {
                                            "id": "3fa85f64-5717-4562-b3fc-2c963f66afa6",
                                            "path": ["$.credentialSubject.id"],
                                        }
                                    ],
                                    "is_holder": [
                                        {
                                            "directive": "required",
                                            "field_id": [
                                                "3fa85f64-5717-4562-b3fc-2c963f66afa6"
                                            ],
                                        }
                                    ],
                                    "limit_disclosure": "required",
                                },
                                "id": "XXXXXXX",
                                "name": "XXXXXXX",
                            }
                        ],
                    },
                }
            },
        }
        with self.assertRaises(TypeError):
            test_module.V20PresSendRequestRequestSchema.load(test_input)



"""Settings implementation for plugins."""

from typing import Any, Mapping, Optional

from .base import BaseSettings

PLUGIN_CONFIG_KEY = "plugin_config"


class PluginSettings(BaseSettings):
    """Retrieve immutable settings for plugins.

    Plugin settings should be retrieved by calling:

        PluginSettings.for_plugin(settings, "my_plugin", {"default": "values"})

    This will extract the PLUGIN_CONFIG_KEY in "settings" and return a new
    PluginSettings instance.
    """

    def __init__(self, values: Optional[Mapping[str, Any]] = None):
        """Initialize a Settings object.

        Args:
            values: An optional dictionary of settings
        """
        self._values = {}
        if values:
            self._values.update(values)

    def __contains__(self, index):
        """Define 'in' operator."""
        return index in self._values

    def __iter__(self):
        """Iterate settings keys."""
        return iter(self._values)

    def __len__(self):
        """Fetch the length of the mapping."""
        return len(self._values)

    def __bool__(self):
        """Convert settings to a boolean."""
        return True

    def copy(self) -> BaseSettings:
        """Produce a copy of the settings instance."""
        return PluginSettings(self._values)

    def extend(self, other: Mapping[str, Any]) -> BaseSettings:
        """Merge another settings instance to produce a new instance."""
        vals = self._values.copy()
        vals.update(other)
        return PluginSettings(vals)

    def to_dict(self) -> dict:
        """Return a dict of the settings instance."""
        setting_dict = {}
        for k in self:
            setting_dict[k] = self[k]
        return setting_dict

    def get_value(self, *var_names: str, default: Optional[Any] = None):
        """Fetch a setting.

        Args:
            var_names: A list of variable name alternatives
            default: The default value to return if none are defined
        """
        for k in var_names:
            if k in self._values:
                return self._values[k]
        return default

    @classmethod
    def for_plugin(
        cls,
        settings: BaseSettings,
        plugin: str,
        default: Optional[Mapping[str, Any]] = None,
    ) -> "PluginSettings":
        """Construct a PluginSettings object from another settings object.

        PLUGIN_CONFIG_KEY is read from settings.
        """
        return cls(settings.get(PLUGIN_CONFIG_KEY, {}).get(plugin, default))



"""Resolve did document admin routes."""

from aiohttp import web
from aiohttp_apispec import docs, match_info_schema, response_schema
from marshmallow import fields, validate
from pydid.common import DID_PATTERN

from ..admin.decorators.auth import tenant_authentication
from ..admin.request_context import AdminRequestContext
from ..messaging.models.openapi import OpenAPISchema
from .base import DIDMethodNotSupported, DIDNotFound, ResolutionResult, ResolverError
from .did_resolver import DIDResolver


class ResolutionResultSchema(OpenAPISchema):
    """Result schema for did document query."""

    did_document = fields.Dict(required=True, metadata={"description": "DID Document"})
    metadata = fields.Dict(required=True, metadata={"description": "Resolution metadata"})


class W3cDID(validate.Regexp):
    """Validate value against w3c DID."""

    EXAMPLE = "did:ted:WgWxqztrNooG92RXvxSTWv"
    PATTERN = DID_PATTERN

    def __init__(self):
        """Initialize the instance."""

        super().__init__(
            W3cDID.PATTERN,
            error="Value {input} is not a w3c decentralized identifier (DID)",
        )


class DIDMatchInfoSchema(OpenAPISchema):
    """Path parameters and validators for request taking DID."""

    did = fields.Str(
        required=True,
        validate=W3cDID(),
        metadata={"description": "DID", "example": W3cDID.EXAMPLE},
    )


@docs(tags=["resolver"], summary="Retrieve doc for requested did")
@match_info_schema(DIDMatchInfoSchema())
@response_schema(ResolutionResultSchema(), 200)
@tenant_authentication
async def resolve_did(request: web.Request):
    """Retrieve a did document."""
    context: AdminRequestContext = request["context"]

    did = request.match_info["did"]
    try:
        async with context.profile.session() as session:
            resolver = session.inject(DIDResolver)
        result: ResolutionResult = await resolver.resolve_with_metadata(
            context.profile, did
        )
    except DIDNotFound as err:
        raise web.HTTPNotFound(reason=err.roll_up) from err
    except DIDMethodNotSupported as err:
        raise web.HTTPNotImplemented(reason=err.roll_up) from err
    except ResolverError as err:
        raise web.HTTPInternalServerError(reason=err.roll_up) from err
    return web.json_response(result.serialize())


async def register(app: web.Application):
    """Register routes."""

    app.add_routes(
        [
            web.get(
                "/resolver/resolve/{did}",
                resolve_did,
                allow_head=False,
            ),
        ]
    )


def post_process_routes(app: web.Application):
    """Amend swagger API."""

    # Add top-level tags description
    if "tags" not in app._state["swagger_dict"]:
        app._state["swagger_dict"]["tags"] = []
    app._state["swagger_dict"]["tags"].append(
        {
            "name": "resolver",
            "description": "did resolver interface.",
            "externalDocs": {
                "description": "DID Resolution Specification",
                "url": "https://www.w3.org/TR/did-core/#resolution",
            },
        }
    )



from unittest import IsolatedAsyncioTestCase

from ......core.oob_processor import OobMessageProcessor
from ......messaging.request_context import RequestContext
from ......messaging.responder import MockResponder
from ......tests import mock
from ......transport.inbound.receipt import MessageReceipt
from ......utils.testing import create_test_profile
from ...messages.pres import V20Pres
from .. import pres_handler as test_module


class TestV20PresHandler(IsolatedAsyncioTestCase):
    async def test_called(self):
        request_context = RequestContext.test_context(await create_test_profile())
        request_context.message_receipt = MessageReceipt()
        request_context.settings["debug.auto_verify_presentation"] = False

        oob_record = mock.MagicMock()
        mock_oob_processor = mock.MagicMock(OobMessageProcessor, autospec=True)
        mock_oob_processor.find_oob_record_for_inbound_message = mock.CoroutineMock(
            return_value=oob_record
        )
        request_context.injector.bind_instance(OobMessageProcessor, mock_oob_processor)

        with mock.patch.object(
            test_module, "V20PresManager", autospec=True
        ) as mock_pres_mgr:
            mock_pres_mgr.return_value.receive_pres = mock.CoroutineMock()
            request_context.message = V20Pres()
            request_context.connection_ready = True
            request_context.connection_record = mock.MagicMock()
            handler = test_module.V20PresHandler()
            responder = MockResponder()
            await handler.handle(request_context, responder)

        mock_pres_mgr.assert_called_once_with(request_context.profile)
        mock_pres_mgr.return_value.receive_pres.assert_called_once_with(
            request_context.message, request_context.connection_record, oob_record
        )
        assert not responder.messages

    async def test_called_auto_verify(self):
        request_context = RequestContext.test_context(await create_test_profile())
        request_context.message_receipt = MessageReceipt()
        request_context.settings["debug.auto_verify_presentation"] = True

        oob_record = mock.MagicMock()
        mock_oob_processor = mock.MagicMock(OobMessageProcessor, autospec=True)
        mock_oob_processor.find_oob_record_for_inbound_message = mock.CoroutineMock(
            return_value=oob_record
        )
        request_context.injector.bind_instance(OobMessageProcessor, mock_oob_processor)

        with mock.patch.object(
            test_module, "V20PresManager", autospec=True
        ) as mock_pres_mgr:
            mock_pres_mgr.return_value.receive_pres = mock.CoroutineMock()
            mock_pres_mgr.return_value.verify_pres = mock.CoroutineMock()
            request_context.message = V20Pres()
            request_context.connection_ready = True
            request_context.connection_record = mock.MagicMock()
            handler = test_module.V20PresHandler()
            responder = MockResponder()
            await handler.handle(request_context, responder)

        mock_pres_mgr.assert_called_once_with(request_context.profile)
        mock_pres_mgr.return_value.receive_pres.assert_called_once_with(
            request_context.message, request_context.connection_record, oob_record
        )
        assert not responder.messages

    async def test_called_auto_verify_x(self):
        request_context = RequestContext.test_context(await create_test_profile())
        request_context.message_receipt = MessageReceipt()
        request_context.settings["debug.auto_verify_presentation"] = True

        oob_record = mock.MagicMock()
        mock_oob_processor = mock.MagicMock(OobMessageProcessor, autospec=True)
        mock_oob_processor.find_oob_record_for_inbound_message = mock.CoroutineMock(
            return_value=oob_record
        )
        request_context.injector.bind_instance(OobMessageProcessor, mock_oob_processor)

        with mock.patch.object(
            test_module, "V20PresManager", autospec=True
        ) as mock_pres_mgr:
            mock_pres_mgr.return_value = mock.MagicMock(
                receive_pres=mock.CoroutineMock(
                    return_value=mock.MagicMock(save_error_state=mock.CoroutineMock())
                ),
                verify_pres=mock.CoroutineMock(side_effect=test_module.LedgerError()),
            )

            request_context.message = V20Pres()
            request_context.connection_ready = True
            request_context.connection_record = mock.MagicMock()
            handler = test_module.V20PresHandler()
            responder = MockResponder()

            with mock.patch.object(
                handler._logger, "exception", mock.MagicMock()
            ) as mock_log_exc:
                await handler.handle(request_context, responder)
                mock_log_exc.assert_called_once()



from unittest import IsolatedAsyncioTestCase

from ......storage.error import StorageDuplicateError, StorageNotFoundError
from ......tests import mock
from ......utils.testing import create_test_profile
from .....didcomm_prefix import DIDCommPrefix
from ...messages.disclosures import Disclosures
from ...messages.queries import Queries, QueryItem
from ..discovery_record import V20DiscoveryExchangeRecord


class TestV20DiscoveryExchangeRecord(IsolatedAsyncioTestCase):
    """Test de/serialization."""

    async def test_record(self):
        same = [
            V20DiscoveryExchangeRecord(
                discovery_exchange_id="3fa85f64-5717-4562-b3fc-2c963f66afa6",
                queries_msg=Queries(
                    queries=[
                        QueryItem(feature_type="protocol", match="*"),
                        QueryItem(feature_type="goal-code", match="test"),
                    ]
                ),
                disclosures=Disclosures(
                    disclosures=[
                        {
                            "id": DIDCommPrefix.qualify_current(
                                "basicmessage/1.0/message"
                            ),
                            "feature-type": "protocol",
                            "roles": [],
                        }
                    ]
                ),
            )
        ] * 2
        diff = [
            V20DiscoveryExchangeRecord(
                discovery_exchange_id="3fa85f64-5717-4562-b3fc-2c963f66afa6",
                queries_msg=Queries(
                    queries=[
                        QueryItem(feature_type="protocol", match="test1.*"),
                        QueryItem(feature_type="goal-code", match="test1"),
                    ]
                ),
                disclosures=Disclosures(
                    disclosures=[
                        {
                            "id": DIDCommPrefix.qualify_current(
                                "basicmessage/1.0/message"
                            ),
                            "feature-type": "protocol",
                            "roles": [],
                        }
                    ]
                ),
            ),
            V20DiscoveryExchangeRecord(
                discovery_exchange_id="3fa85f64-5717-4562-b3fc-2c963f66afa6",
                queries_msg=Queries(
                    queries=[
                        QueryItem(feature_type="protocol", match="test2.*"),
                        QueryItem(feature_type="goal-code", match="test2"),
                    ]
                ),
            ),
        ]

        for i in range(len(same) - 1):
            for j in range(i, len(same)):
                assert same[i] == same[j]

        for i in range(len(diff) - 1):
            for j in range(i, len(diff)):
                assert diff[i] == diff[j] if i == j else diff[i] != diff[j]

    async def test_serde(self):
        """Test de/serialization."""
        queries = Queries(
            queries=[
                QueryItem(feature_type="protocol", match="*"),
                QueryItem(feature_type="goal-code", match="test"),
            ]
        )
        disclosures = Disclosures(
            disclosures=[
                {
                    "id": DIDCommPrefix.qualify_current("basicmessage/1.0/message"),
                    "feature-type": "protocol",
                    "roles": [],
                }
            ]
        )
        ex_rec = V20DiscoveryExchangeRecord(
            discovery_exchange_id="3fa85f64-5717-4562-b3fc-2c963f66afa6",
            disclosures=disclosures,
        )
        ex_rec.queries_msg = queries
        assert isinstance(ex_rec.queries_msg, Queries)
        ser = ex_rec.serialize()
        deser = V20DiscoveryExchangeRecord.deserialize(ser)
        assert isinstance(deser.queries_msg, Queries)

        assert isinstance(ex_rec.disclosures, Disclosures)
        ser = ex_rec.serialize()
        deser = V20DiscoveryExchangeRecord.deserialize(ser)
        assert isinstance(deser.disclosures, Disclosures)

    async def test_retrieve_by_conn_id(self):
        self.profile = await create_test_profile()
        async with self.profile.session() as session:
            record = V20DiscoveryExchangeRecord(
                queries_msg=Queries(
                    queries=[
                        QueryItem(feature_type="protocol", match="*"),
                        QueryItem(feature_type="goal-code", match="test"),
                    ],
                ),
                connection_id="test123",
            )
            await record.save(session)
            retrieved = await V20DiscoveryExchangeRecord.retrieve_by_connection_id(
                session=session, connection_id="test123"
            )
            assert retrieved
            assert retrieved.connection_id == "test123"

    async def test_exists_for_connection_id(self):
        self.profile = await create_test_profile()
        async with self.profile.session() as session:
            record = V20DiscoveryExchangeRecord(
                queries_msg=Queries(
                    queries=[
                        QueryItem(feature_type="protocol", match="*"),
                        QueryItem(feature_type="goal-code", match="test"),
                    ],
                ),
                connection_id="test123",
            )
            await record.save(session)
            check = await V20DiscoveryExchangeRecord.exists_for_connection_id(
                session=session, connection_id="test123"
            )
            assert check

    async def test_exists_for_connection_id_not_found(self):
        self.profile = await create_test_profile()
        async with self.profile.session() as session:
            with mock.patch.object(
                V20DiscoveryExchangeRecord,
                "retrieve_by_tag_filter",
                mock.CoroutineMock(),
            ) as mock_retrieve_by_tag_filter:
                mock_retrieve_by_tag_filter.side_effect = StorageNotFoundError
                check = await V20DiscoveryExchangeRecord.exists_for_connection_id(
                    session=session, connection_id="test123"
                )
                assert not check

    async def test_exists_for_connection_id_duplicate(self):
        self.profile = await create_test_profile()
        async with self.profile.session() as session:
            with mock.patch.object(
                V20DiscoveryExchangeRecord,
                "retrieve_by_tag_filter",
                mock.CoroutineMock(),
            ) as mock_retrieve_by_tag_filter:
                mock_retrieve_by_tag_filter.side_effect = StorageDuplicateError
                check = await V20DiscoveryExchangeRecord.exists_for_connection_id(
                    session=session, connection_id="test123"
                )
                assert check



"""AnonCreds cred def OpenAPI validators."""

from typing import Any, Dict, List, Optional

from anoncreds import RevocationRegistryDefinition, RevocationStatusList
from marshmallow import EXCLUDE, fields
from marshmallow.validate import OneOf
from typing_extensions import Literal

from ...messaging.models.base import BaseModel, BaseModelSchema
from ...messaging.valid import (
    ANONCREDS_CRED_DEF_ID_EXAMPLE,
    ANONCREDS_DID_EXAMPLE,
    ANONCREDS_REV_REG_ID_EXAMPLE,
    ISO8601_DATETIME_EXAMPLE,
    RAW_ED25519_2018_PUBLIC_KEY_EXAMPLE,
)


class RevRegDefValue(BaseModel):
    """RevRegDefValue model."""

    class Meta:
        """RevRegDefValue metadata."""

        schema_class = "RevRegDefValueSchema"

    def __init__(
        self,
        public_keys: dict,
        max_cred_num: int,
        tails_location: str,
        tails_hash: str,
        **kwargs,
    ):
        """Initialize an instance.

        Args:
            public_keys: Public Keys
            max_cred_num: Max. number of Creds
            tails_location: Tails file location
            tails_hash: Tails file hash
            kwargs: Keyword arguments
        """
        super().__init__(**kwargs)
        self.public_keys = public_keys
        self.max_cred_num = max_cred_num
        self.tails_location = tails_location
        self.tails_hash = tails_hash


class RevRegDefValueSchema(BaseModelSchema):
    """RevRegDefValue schema."""

    class Meta:
        """RevRegDefValueSchema metadata."""

        model_class = RevRegDefValue
        unknown = EXCLUDE

    public_keys = fields.Dict(
        data_key="publicKeys", metadata={"example": RAW_ED25519_2018_PUBLIC_KEY_EXAMPLE}
    )
    max_cred_num = fields.Int(data_key="maxCredNum", metadata={"example": 777})
    tails_location = fields.Str(
        data_key="tailsLocation",
        metadata={
            "example": "https://tails-server.com/hash/7Qen9RDyemMuV7xGQvp7NjwMSpyHieJyBakycxN7dX7P"
        },
    )
    tails_hash = fields.Str(
        data_key="tailsHash",
        metadata={"example": "7Qen9RDyemMuV7xGQvp7NjwMSpyHieJyBakycxN7dX7P"},
    )


class RevRegDef(BaseModel):
    """RevRegDef."""

    class Meta:
        """RevRegDef metadata."""

        schema_class = "RevRegDefSchema"

    def __init__(
        self,
        issuer_id: str,
        type: Literal["CL_ACCUM"],
        cred_def_id: str,
        tag: str,
        value: RevRegDefValue,
        **kwargs,
    ):
        """Initialize an instance.

        Args:
            issuer_id: Issuer ID
            type: type
            cred_def_id: Cred Def ID
            tag: Tag
            value: Rev Reg Def Value
            kwargs: Keyword arguments
        """
        super().__init__(**kwargs)
        self.issuer_id = issuer_id
        self.type = type
        self.cred_def_id = cred_def_id
        self.tag = tag
        self.value = value

    @classmethod
    def from_native(cls, rev_reg_def: RevocationRegistryDefinition):
        """Convert a native revocation registry definition to a RevRegDef object."""
        return cls.deserialize(rev_reg_def.to_json())

    def to_native(self) -> RevocationRegistryDefinition:
        """Convert to native anoncreds revocation registry definition."""
        return RevocationRegistryDefinition.load(self.serialize())


class RevRegDefSchema(BaseModelSchema):
    """RevRegDefSchema."""

    class Meta:
        """RevRegDefSchema metadata."""

        model_class = RevRegDef
        unknown = EXCLUDE

    issuer_id = fields.Str(
        metadata={
            "description": "Issuer Identifier of the credential definition or schema",
            "example": ANONCREDS_DID_EXAMPLE,
        },
        data_key="issuerId",
    )
    type = fields.Str(data_key="revocDefType")
    cred_def_id = fields.Str(
        metadata={
            "description": "Credential definition identifier",
            "example": ANONCREDS_CRED_DEF_ID_EXAMPLE,
        },
        data_key="credDefId",
    )
    tag = fields.Str(
        metadata={
            "description": "tag for the revocation registry definition",
            "example": "default",
        }
    )
    value = fields.Nested(RevRegDefValueSchema())


class RevRegDefState(BaseModel):
    """RevRegDefState."""

    STATE_FINISHED = "finished"
    STATE_FAILED = "failed"
    STATE_ACTION = "action"
    STATE_WAIT = "wait"
    STATE_DECOMMISSIONED = "decommissioned"
    STATE_FULL = "full"

    class Meta:
        """RevRegDefState metadata."""

        schema_class = "RevRegDefStateSchema"

    def __init__(
        self,
        state: str,
        revocation_registry_definition_id: str,
        revocation_registry_definition: RevRegDef,
    ):
        """Initialize an instance.

        Args:
            state: State
            revocation_registry_definition_id: Rev Reg Definition ID
            revocation_registry_definition: Rev Reg Definition
        """
        self.state = state
        self.revocation_registry_definition_id = revocation_registry_definition_id
        self.revocation_registry_definition = revocation_registry_definition


class RevRegDefStateSchema(BaseModelSchema):
    """RevRegDefStateSchema."""

    class Meta:
        """RevRegDefStateSchema metadata."""

        model_class = RevRegDefState
        unknown = EXCLUDE

    state = fields.Str(
        validate=OneOf(
            [
                RevRegDefState.STATE_FINISHED,
                RevRegDefState.STATE_FAILED,
                RevRegDefState.STATE_ACTION,
                RevRegDefState.STATE_WAIT,
                RevRegDefState.STATE_DECOMMISSIONED,
                RevRegDefState.STATE_FULL,
            ]
        )
    )
    revocation_registry_definition_id = fields.Str(
        metadata={
            "description": "revocation registry definition id",
            "example": ANONCREDS_REV_REG_ID_EXAMPLE,
        }
    )
    revocation_registry_definition = fields.Nested(
        RevRegDefSchema(), metadata={"description": "revocation registry definition"}
    )


class RevRegDefResult(BaseModel):
    """Cred def result."""

    class Meta:
        """RevRegDefResult metadata."""

        schema_class = "RevRegDefResultSchema"

    def __init__(
        self,
        job_id: Optional[str],
        revocation_registry_definition_state: RevRegDefState,
        registration_metadata: dict,
        revocation_registry_definition_metadata: dict,
        **kwargs,
    ):
        """Initialize an instance.

        Args:
            job_id: Job ID
            revocation_registry_definition_state: Rev Reg Def state
            registration_metadata: Registration metadata
            revocation_registry_definition_metadata: Rev Reg Def metadata
            kwargs: Keyword arguments
        """
        super().__init__(**kwargs)
        self.job_id = job_id
        self.revocation_registry_definition_state = revocation_registry_definition_state
        self.registration_metadata = registration_metadata
        self.revocation_registry_definition_metadata = (
            revocation_registry_definition_metadata
        )

    @property
    def rev_reg_def_id(self) -> str:
        """Revocation Registry Definition ID."""
        return self.revocation_registry_definition_state.revocation_registry_definition_id

    @property
    def rev_reg_def(self) -> RevRegDef:
        """Revocation Registry Definition."""
        return self.revocation_registry_definition_state.revocation_registry_definition


class RevRegDefResultSchema(BaseModelSchema):
    """Cred def result schema."""

    class Meta:
        """RevRegDefResultSchema metadata."""

        model_class = RevRegDefResult
        unknown = EXCLUDE

    job_id = fields.Str()
    revocation_registry_definition_state = fields.Nested(RevRegDefStateSchema())
    registration_metadata = fields.Dict()
    # For indy, revocation_registry_definition_metadata will contain the seqNo
    revocation_registry_definition_metadata = fields.Dict()


class GetRevRegDefResult(BaseModel):
    """GetRevRegDefResult."""

    class Meta:
        """GetRevRegDefResult metadata."""

        schema_class = "GetRevRegDefResultSchema"

    def __init__(
        self,
        revocation_registry: RevRegDef,
        revocation_registry_id: str,
        resolution_metadata: Dict[str, Any],
        revocation_registry_metadata: Dict[str, Any],
        **kwargs,
    ):
        """Initialize an instance.

        Args:
            revocation_registry: Revocation registry
            revocation_registry_id: Revocation Registry ID
            resolution_metadata: Resolution metadata
            revocation_registry_metadata: Revocation Registry metadata
            kwargs: Keyword arguments
        """
        super().__init__(**kwargs)
        self.revocation_registry = revocation_registry
        self.revocation_registry_id = revocation_registry_id
        self.resolution_metadata = resolution_metadata
        self.revocation_registry_metadata = revocation_registry_metadata


class GetRevRegDefResultSchema(BaseModelSchema):
    """GetRevRegDefResultSchema."""

    class Meta:
        """GetRevRegDefResultSchema metadata."""

        model_class = GetRevRegDefResult
        unknown = EXCLUDE

    revocation_registry = fields.Nested(RevRegDefSchema())
    revocation_registry_id = fields.Str()
    resolution_metadata = fields.Dict()
    revocation_registry_metadata = fields.Dict()


class RevList(BaseModel):
    """RevList."""

    class Meta:
        """RevList metadata."""

        schema_class = "RevListSchema"

    def __init__(
        self,
        issuer_id: str,
        rev_reg_def_id: str,
        revocation_list: List[int],
        current_accumulator: str,
        timestamp: Optional[int] = None,
        **kwargs,
    ):
        """Initialize an instance.

        Args:
            issuer_id: Job ID
            rev_reg_def_id: Revocation Registry Def. ID
            revocation_list: Revocation list
            current_accumulator: Current accumulator
            timestamp: Timestamp
            kwargs: Keyword arguments
        """
        super().__init__(**kwargs)
        self.issuer_id = issuer_id
        self.rev_reg_def_id = rev_reg_def_id
        self.revocation_list = revocation_list
        self.current_accumulator = current_accumulator
        self.timestamp = timestamp

    @classmethod
    def from_native(cls, rev_list: RevocationStatusList):
        """Convert from native revocation list."""
        return cls.deserialize(rev_list.to_json())

    def to_native(self) -> RevocationStatusList:
        """Convert to native revocation list."""
        return RevocationStatusList.load(self.serialize())


class RevListSchema(BaseModelSchema):
    """RevListSchema."""

    class Meta:
        """RevListSchema metadata."""

        model_class = RevList
        unknown = EXCLUDE

    issuer_id = fields.Str(
        metadata={
            "description": "Issuer Identifier of the credential definition or schema",
            "example": ANONCREDS_DID_EXAMPLE,
        },
        data_key="issuerId",
    )
    rev_reg_def_id = fields.Str(
        metadata={
            "description": "The ID of the revocation registry definition",
            "example": ANONCREDS_REV_REG_ID_EXAMPLE,
        },
        data_key="revRegDefId",
    )
    revocation_list = fields.List(
        fields.Int(),
        metadata={
            "description": "Bit list representing revoked credentials",
            "example": [0, 1, 1, 0],
        },
        data_key="revocationList",
    )
    current_accumulator = fields.Str(
        metadata={
            "description": "The current accumulator value",
            "example": "21 118...1FB",
        },
        data_key="currentAccumulator",
    )
    timestamp = fields.Int(
        metadata={
            "description": "Timestamp at which revocation list is applicable",
            "example": ISO8601_DATETIME_EXAMPLE,
        },
        required=False,
    )


class RevListState(BaseModel):
    """RevListState."""

    STATE_FINISHED = "finished"
    STATE_FAILED = "failed"
    STATE_ACTION = "action"
    STATE_WAIT = "wait"

    class Meta:
        """RevListState metadata."""

        schema_class = "RevListStateSchema"

    def __init__(
        self,
        state: str,
        revocation_list: RevList,
    ):
        """Initialize an instance.

        Args:
            state: State
            revocation_list: Revocation list
        """
        self.state = state
        self.revocation_list = revocation_list


class RevListStateSchema(BaseModelSchema):
    """RevListStateSchema."""

    class Meta:
        """RevListStateSchema metadata."""

        model_class = RevListState
        unknown = EXCLUDE

    state = fields.Str(
        validate=OneOf(
            [
                RevListState.STATE_FINISHED,
                RevListState.STATE_FAILED,
                RevListState.STATE_ACTION,
                RevListState.STATE_WAIT,
            ]
        )
    )
    revocation_list = fields.Nested(
        RevListSchema(), metadata={"description": "revocation list"}
    )


class RevListResult(BaseModel):
    """Cred def result."""

    class Meta:
        """RevListResult metadata."""

        schema_class = "RevListResultSchema"

    def __init__(
        self,
        job_id: Optional[str],
        revocation_list_state: RevListState,
        registration_metadata: dict,
        revocation_list_metadata: dict,
        **kwargs,
    ):
        """Initialize an instance.

        Args:
            job_id: Job ID
            revocation_list_state: Revocation list state
            registration_metadata: Registration metadata
            revocation_list_metadata: Revocation list metadata
            kwargs: Keyword arguments
        """
        super().__init__(**kwargs)
        self.job_id = job_id
        self.revocation_list_state = revocation_list_state
        self.registration_metadata = registration_metadata
        self.revocation_list_metadata = revocation_list_metadata

    @property
    def rev_reg_def_id(self) -> str:
        """Rev reg def id."""
        return self.revocation_list_state.revocation_list.rev_reg_def_id


class RevListResultSchema(BaseModelSchema):
    """Cred def result schema."""

    class Meta:
        """RevListResultSchema metadata."""

        model_class = RevListResult
        unknown = EXCLUDE

    job_id = fields.Str()
    revocation_list_state = fields.Nested(RevListStateSchema())
    registration_metadata = fields.Dict()
    # For indy, revocation_list_metadata will contain the seqNo
    revocation_list_metadata = fields.Dict()


class GetRevListResult(BaseModel):
    """GetRevListResult."""

    class Meta:
        """GetRevListResult metadata."""

        schema_class = "GetRevListResultSchema"

    def __init__(
        self,
        revocation_list: RevList,
        resolution_metadata: Dict[str, Any],
        revocation_registry_metadata: Dict[str, Any],
        **kwargs,
    ):
        """Initialize an instance.

        Args:
            revocation_list: Revocation list
            resolution_metadata: Resolution metadata
            revocation_registry_metadata: Rev Reg metadata
            kwargs: Keyword arguments
        """
        super().__init__(**kwargs)
        self.revocation_list = revocation_list
        self.resolution_metadata = resolution_metadata
        self.revocation_registry_metadata = revocation_registry_metadata


class GetRevListResultSchema(BaseModelSchema):
    """GetRevListResultSchema."""

    class Meta:
        """GetRevListResultSchema metadata."""

        model_class = GetRevListResult
        unknown = EXCLUDE

    revocation_list = fields.Nested(RevListSchema)
    resolution_metadata = fields.Str()
    revocation_registry_metadata = fields.Dict()



"""Indy revocation registry management."""

import logging
from typing import Optional, Sequence, Tuple

from uuid_utils import uuid4

from ..core.profile import Profile
from ..ledger.base import BaseLedger
from ..ledger.multiple_ledger.ledger_requests_executor import (
    GET_CRED_DEF,
    GET_REVOC_REG_DEF,
    IndyLedgerRequestsExecutor,
)
from ..multitenant.base import BaseMultitenantManager
from ..protocols.endorse_transaction.v1_0.util import (
    get_endorser_connection_id,
    is_author_role,
)
from ..storage.base import StorageNotFoundError
from ..wallet.askar import CATEGORY_DID
from ..wallet.error import WalletNotFoundError
from .error import (
    RevocationError,
    RevocationInvalidStateValueError,
    RevocationNotSupportedError,
    RevocationRegistryBadSizeError,
)
from .models.issuer_rev_reg_record import IssuerRevRegRecord
from .models.revocation_registry import RevocationRegistry
from .util import notify_revocation_reg_init_event

LOGGER = logging.getLogger(__name__)


class IndyRevocation:
    """Class for managing Indy credential revocation."""

    REV_REG_CACHE = {}

    def __init__(self, profile: Profile):
        """Initialize the IndyRevocation instance."""
        self._profile = profile

    async def init_issuer_registry(
        self,
        cred_def_id: str,
        max_cred_num: Optional[int] = None,
        revoc_def_type: Optional[str] = None,
        tag: Optional[str] = None,
        create_pending_rev_reg: bool = False,
        endorser_connection_id: Optional[str] = None,
        notify: bool = True,
    ) -> IssuerRevRegRecord:
        """Create a new revocation registry record for a credential definition."""
        multitenant_mgr = self._profile.inject_or(BaseMultitenantManager)
        if multitenant_mgr:
            ledger_exec_inst = IndyLedgerRequestsExecutor(self._profile)
        else:
            ledger_exec_inst = self._profile.inject(IndyLedgerRequestsExecutor)
        ledger = (
            await ledger_exec_inst.get_ledger_for_identifier(
                cred_def_id,
                txn_record_type=GET_CRED_DEF,
            )
        )[1]
        async with ledger:
            cred_def = await ledger.get_credential_definition(cred_def_id)
        if not cred_def:
            raise RevocationNotSupportedError("Credential definition not found")
        if not cred_def["value"].get("revocation"):
            raise RevocationNotSupportedError(
                "Credential definition does not support revocation"
            )
        if max_cred_num and not (
            RevocationRegistry.MIN_SIZE <= max_cred_num <= RevocationRegistry.MAX_SIZE
        ):
            raise RevocationRegistryBadSizeError(
                f"Bad revocation registry size: {max_cred_num}"
            )

        record_id = str(uuid4())
        issuer_did = cred_def_id.split(":")[0]
        # Try and get a did:indy did from nym value stored as a did
        async with self._profile.session() as session:
            try:
                indy_did = await session.handle.fetch(
                    CATEGORY_DID, f"did:indy:{issuer_did}"
                )
            except WalletNotFoundError:
                indy_did = None

        record = IssuerRevRegRecord(
            new_with_id=True,
            record_id=record_id,
            cred_def_id=cred_def_id,
            issuer_did=indy_did.name if indy_did else issuer_did,
            max_cred_num=max_cred_num,
            revoc_def_type=revoc_def_type,
            tag=tag,
        )
        revoc_def_type = record.revoc_def_type
        rtag = record.tag or record_id
        record.revoc_reg_id = f"{issuer_did}:4:{cred_def_id}:{revoc_def_type}:{rtag}"
        async with self._profile.session() as session:
            await record.save(session, reason="Init revocation registry")

        if endorser_connection_id is None and is_author_role(self._profile):
            endorser_connection_id = await get_endorser_connection_id(self._profile)
            if not endorser_connection_id:
                raise RevocationError(reason="Endorser connection not found")

        if notify:
            await notify_revocation_reg_init_event(
                self._profile,
                record.record_id,
                create_pending_rev_reg=create_pending_rev_reg,
                endorser_connection_id=endorser_connection_id,
            )

        return record

    async def handle_full_registry(self, revoc_reg_id: str):
        """Update the registry status and start the next registry generation."""
        await self._set_registry_status(revoc_reg_id, IssuerRevRegRecord.STATE_FULL)

    async def decommission_registry(self, cred_def_id: str):
        """Decommission post-init registries and start the next registry generation."""
        async with self._profile.session() as session:
            registries = await IssuerRevRegRecord.query_by_cred_def_id(
                session, cred_def_id
            )

        # decommission everything except init
        recs = list(
            filter(lambda r: r.state != IssuerRevRegRecord.STATE_INIT, registries)
        )

        for rec in recs:
            LOGGER.debug(f"decommission {rec.state} rev. reg.")
            LOGGER.debug(f"revoc_reg_id: {rec.revoc_reg_id}")
            LOGGER.debug(f"cred_def_id: {cred_def_id}")
            # decommission active registry, we need to init a replacement
            init = IssuerRevRegRecord.STATE_ACTIVE == rec.state
            await self._set_registry_status(
                rec.revoc_reg_id, IssuerRevRegRecord.STATE_DECOMMISSIONED, init
            )

        return recs

    async def _set_registry_status(
        self, revoc_reg_id: str, state: str, init: bool = True
    ):
        """Update the registry status and start the next registry generation."""
        if state not in IssuerRevRegRecord.STATES:
            raise RevocationInvalidStateValueError(
                reason=f"{state} is not a valid Revocation Registry state value."
            )
        async with self._profile.transaction() as txn:
            registry = await IssuerRevRegRecord.retrieve_by_revoc_reg_id(
                txn, revoc_reg_id, for_update=True
            )
            if registry.state == state:
                return
            await registry.set_state(
                txn,
                state,
            )
            await txn.commit()

        if (state in IssuerRevRegRecord.TERMINAL_STATES) and init:
            return await self.init_issuer_registry(
                registry.cred_def_id,
                registry.max_cred_num,
                registry.revoc_def_type,
            )

        return

    async def get_active_issuer_rev_reg_record(
        self, cred_def_id: str
    ) -> IssuerRevRegRecord:
        """Return current active registry for issuing a given credential definition.

        Args:
            cred_def_id: ID of the base credential definition
        """
        async with self._profile.session() as session:
            current = sorted(
                await IssuerRevRegRecord.query_by_cred_def_id(
                    session, cred_def_id, IssuerRevRegRecord.STATE_ACTIVE
                )
            )
        if current:
            return current[0]  # active record is oldest published but not full
        raise StorageNotFoundError(
            f"No active issuer revocation record found for cred def id {cred_def_id}"
        )

    async def get_issuer_rev_reg_record(self, revoc_reg_id: str) -> IssuerRevRegRecord:
        """Return a revocation registry record by identifier.

        Args:
            revoc_reg_id: ID of the revocation registry
        """
        async with self._profile.session() as session:
            return await IssuerRevRegRecord.retrieve_by_revoc_reg_id(
                session, revoc_reg_id
            )

    async def list_issuer_registries(self) -> Sequence[IssuerRevRegRecord]:
        """List the issuer's current revocation registries."""
        async with self._profile.session() as session:
            return await IssuerRevRegRecord.query(session)

    async def get_issuer_rev_reg_delta(
        self,
        rev_reg_id: str,
        timestamp_from: Optional[int] = None,
        timestamp_to: Optional[int] = None,
    ) -> dict:
        """Check ledger for revocation status for a given revocation registry.

        Args:
            rev_reg_id (str): ID of the revocation registry
            timestamp_from (int, optional): The sequence number to start from (exclusive).
                Defaults to None.
            timestamp_to (int, optional): The sequence number to end at (inclusive).
                Defaults to None.

        Returns:
            dict: The revocation registry delta.

        """
        ledger = await self.get_ledger_for_registry(rev_reg_id)
        async with ledger:
            (rev_reg_delta, _) = await ledger.get_revoc_reg_delta(
                rev_reg_id,
                timestamp_from,
                timestamp_to,
            )

        return rev_reg_delta

    async def get_or_create_active_registry(
        self, cred_def_id: str
    ) -> Optional[Tuple[IssuerRevRegRecord, RevocationRegistry]]:
        """Fetch the active revocation registry.

        If there is no active registry then creation of a new registry will be
        triggered and the caller should retry after a delay.
        """
        try:
            active_rev_reg_rec = await self.get_active_issuer_rev_reg_record(cred_def_id)
            rev_reg = active_rev_reg_rec.get_registry()
            await rev_reg.get_or_fetch_local_tails_path()
            return active_rev_reg_rec, rev_reg
        except StorageNotFoundError:
            pass

        async with self._profile.session() as session:
            full_registries = await IssuerRevRegRecord.query_by_cred_def_id(
                session, cred_def_id, None, IssuerRevRegRecord.STATE_FULL, 1
            )

            # all registries are full, create a new one
            if not full_registries:
                # Use any registry to get max cred num
                any_registry = await IssuerRevRegRecord.query_by_cred_def_id(
                    session, cred_def_id, limit=1
                )
                if not any_registry:
                    raise RevocationError(
                        f"No revocation registry record found in issuer wallet for cred def id {cred_def_id}"  # noqa: E501
                    )
                any_registry = any_registry[0]
                await self.init_issuer_registry(
                    cred_def_id,
                    max_cred_num=any_registry.max_cred_num,
                )
            # if there is a posted registry, activate oldest
            else:
                posted_registries = await IssuerRevRegRecord.query_by_cred_def_id(
                    session, cred_def_id, IssuerRevRegRecord.STATE_POSTED, None, None
                )
                if posted_registries:
                    posted_registries = sorted(
                        posted_registries, key=lambda r: r.created_at
                    )
                    await self._set_registry_status(
                        revoc_reg_id=posted_registries[0].revoc_reg_id,
                        state=IssuerRevRegRecord.STATE_ACTIVE,
                    )
        return None

    async def get_ledger_registry(self, revoc_reg_id: str) -> RevocationRegistry:
        """Get a revocation registry from the ledger, fetching as necessary."""
        if revoc_reg_id in IndyRevocation.REV_REG_CACHE:
            return IndyRevocation.REV_REG_CACHE[revoc_reg_id]

        ledger = await self.get_ledger_for_registry(revoc_reg_id)

        async with ledger:
            rev_reg = RevocationRegistry.from_definition(
                await ledger.get_revoc_reg_def(revoc_reg_id), True
            )
            IndyRevocation.REV_REG_CACHE[revoc_reg_id] = rev_reg
            return rev_reg

    async def get_ledger_for_registry(self, revoc_reg_id: str) -> BaseLedger:
        """Get the ledger for the given registry."""
        multitenant_mgr = self._profile.inject_or(BaseMultitenantManager)
        if multitenant_mgr:
            ledger_exec_inst = IndyLedgerRequestsExecutor(self._profile)
        else:
            ledger_exec_inst = self._profile.inject(IndyLedgerRequestsExecutor)
        ledger = (
            await ledger_exec_inst.get_ledger_for_identifier(
                revoc_reg_id,
                txn_record_type=GET_REVOC_REG_DEF,
            )
        )[1]
        return ledger



from unittest import TestCase

from ...core.error import BaseError
from ..did_method import KEY, SOV, DIDMethods
from ..key_type import BLS12381G2, ED25519

SOV_DID_METHOD_NAME = "sov"
SOV_SUPPORTED_KEY_TYPES = [ED25519]
KEY_DID_METHOD_NAME = "key"


class TestDidMethod(TestCase):
    """TestCases for did method"""

    did_methods = DIDMethods()

    def test_from_metadata(self):
        """Testing 'from_metadata'"""
        assert self.did_methods.from_metadata({"method": SOV_DID_METHOD_NAME}) == SOV
        assert self.did_methods.from_metadata({"method": KEY_DID_METHOD_NAME}) == KEY

        # test backwards compat
        assert self.did_methods.from_metadata({}) == SOV

    def test_from_method(self):
        """Testing 'from_method'"""
        assert self.did_methods.from_method(SOV_DID_METHOD_NAME) == SOV
        assert self.did_methods.from_method(KEY_DID_METHOD_NAME) == KEY
        assert self.did_methods.from_method("random") is None

    def test_from_did(self):
        """Testing 'from_did'"""
        assert self.did_methods.from_did(f"did:{SOV_DID_METHOD_NAME}:xxxx") == SOV
        assert self.did_methods.from_did(f"did:{KEY_DID_METHOD_NAME}:xxxx") == KEY

        with self.assertRaises(BaseError) as context:
            self.did_methods.from_did("did:unknown:something")
        assert "Unsupported did method: unknown" in str(context.exception)

    def test_properties(self):
        """Testing 'properties'"""
        method = SOV

        assert method.method_name == SOV_DID_METHOD_NAME
        assert method.supported_key_types == SOV_SUPPORTED_KEY_TYPES
        assert method.supports_rotation is True

        assert method.supports_key_type(ED25519) is True
        assert method.supports_key_type(BLS12381G2) is False



"""Utilities for working with Message Types and Versions."""

import re
from dataclasses import dataclass
from functools import lru_cache
from typing import ClassVar, Pattern, Tuple, Union


@dataclass
class MessageVersion:
    """Message type version."""

    PATTERN: ClassVar[Pattern] = re.compile(r"^(0|[1-9]\d*)\.(0|[1-9]\d*)$")

    major: int
    minor: int

    @classmethod
    @lru_cache
    def from_str(cls, value: str):
        """Parse a version string."""
        if match := cls.PATTERN.match(value):
            return cls(
                int(match.group(1)),
                int(match.group(2)),
            )

        raise ValueError(f"Invalid version: {value}")

    def __gt__(self, other: "MessageVersion") -> bool:
        """Test whether this version is greater than another."""
        if self.major != other.major:
            return self.major > other.major
        return self.minor > other.minor

    def __eq__(self, other: object) -> bool:
        """Equality comparison."""
        if not isinstance(other, MessageVersion):
            return False

        return self.major == other.major and self.minor == other.minor

    def __lt__(self, other: "MessageVersion") -> bool:
        """Test whether this version is less than another."""
        if self.major != other.major:
            return self.major < other.major
        return self.minor < other.minor

    def __str__(self) -> str:
        """Return the version as a string."""
        return f"{self.major}.{self.minor}"

    def __hash__(self) -> int:
        """Return a hash of the version."""
        return hash((self.major, self.minor))

    def compatible(self, other: "MessageVersion") -> bool:
        """Test whether this version is compatible with another."""
        if self == other:
            return True
        return self.major == other.major and self.minor <= other.minor


@dataclass
class ProtocolIdentifier:
    """Protocol identifier."""

    PATTERN: ClassVar[Pattern] = re.compile(r"^(.*?)/([a-z0-9._-]+)/(\d[^/]*)$")
    FROM_MESSAGE_TYPE_PATTERN: ClassVar[Pattern] = re.compile(
        r"^(.*?)/([a-z0-9._-]+)/(\d[^/]*).*$"
    )

    doc_uri: str
    protocol: str
    version: MessageVersion

    @classmethod
    @lru_cache
    def from_str(cls, value: str) -> "ProtocolIdentifier":
        """Parse a protocol identifier string."""
        if match := cls.PATTERN.match(value):
            return cls(
                doc_uri=match.group(1),
                protocol=match.group(2),
                version=MessageVersion.from_str(match.group(3)),
            )
        raise ValueError(f"Invalid protocol identifier: {value}")

    @classmethod
    @lru_cache
    def from_message_type(
        cls, message_type: Union[str, "MessageType"]
    ) -> "ProtocolIdentifier":
        """Create a protocol identifier from a message type."""
        if isinstance(message_type, str):
            if match := cls.FROM_MESSAGE_TYPE_PATTERN.match(message_type):
                return cls(
                    doc_uri=match.group(1),
                    protocol=match.group(2),
                    version=MessageVersion.from_str(match.group(3)),
                )

            raise ValueError(f"Invalid protocol identifier: {message_type}")
        elif isinstance(message_type, MessageType):
            return cls(message_type.doc_uri, message_type.protocol, message_type.version)
        else:
            raise TypeError(f"Invalid message type: {message_type}")

    def __str__(self) -> str:
        """Return the protocol identifier as a string."""
        return f"{self.doc_uri}/{self.protocol}/{self.version}"

    @property
    def stem(self) -> str:
        """Return the protocol stem, including doc_uri, protocol, and major version."""
        return f"{self.doc_uri}/{self.protocol}/{self.version.major}"

    def with_version(
        self, version: Union[str, MessageVersion, Tuple[int, int]]
    ) -> "ProtocolIdentifier":
        """Return a new protocol identifier with the specified version."""
        if isinstance(version, str):
            version = MessageVersion.from_str(version)

        if isinstance(version, tuple):
            version = MessageVersion(*version)

        return ProtocolIdentifier(
            doc_uri=self.doc_uri,
            protocol=self.protocol,
            version=version,
        )


@dataclass
class MessageType:
    """Message type."""

    PATTERN: ClassVar[Pattern] = re.compile(
        r"^(.*?)/([a-z0-9._-]+)/(\d[^/]*)/([a-z0-9._-]+)$"
    )

    doc_uri: str
    protocol: str
    version: MessageVersion
    name: str

    @classmethod
    @lru_cache
    def from_str(cls, value: str):
        """Parse a message type string."""
        if match := cls.PATTERN.match(value):
            return cls(
                doc_uri=match.group(1),
                protocol=match.group(2),
                version=MessageVersion.from_str(match.group(3)),
                name=match.group(4),
            )

        raise ValueError(f"Invalid message type: {value}")

    def __str__(self) -> str:
        """Return the message type as a string."""
        return f"{self.doc_uri}/{self.protocol}/{self.version}/{self.name}"

    def with_version(
        self, version: Union[str, MessageVersion, Tuple[int, int]]
    ) -> "MessageType":
        """Return a new message type with the specified version."""
        if isinstance(version, str):
            version = MessageVersion.from_str(version)

        if isinstance(version, tuple):
            version = MessageVersion(*version)

        return MessageType(
            doc_uri=self.doc_uri,
            protocol=self.protocol,
            version=version,
            name=self.name,
        )

    def __hash__(self) -> int:
        """Return a hash of the message type."""
        return hash((self.doc_uri, self.protocol, self.version, self.name))


class MessageTypeStr(str):
    """Message type string."""

    def __init__(self, value: str):
        """Initialize the message type string."""
        super().__init__()
        self._parsed = MessageType.from_str(value)

    @property
    def parsed(self) -> MessageType:
        """Return the parsed message type."""
        return self._parsed

    @property
    def doc_uri(self) -> str:
        """Return the message type document URI."""
        return self._parsed.doc_uri

    @property
    def protocol(self) -> str:
        """Return the message type protocol."""
        return self._parsed.protocol

    @property
    def version(self) -> MessageVersion:
        """Return the message type version."""
        return self._parsed.version

    @property
    def name(self) -> str:
        """Return the message type name."""
        return self._parsed.name

    @property
    def protocol_identifier(self) -> ProtocolIdentifier:
        """Return the message type protocol identifier."""
        return ProtocolIdentifier.from_message_type(self._parsed)

    def with_version(
        self, version: Union[str, MessageVersion, Tuple[int, int]]
    ) -> "MessageTypeStr":
        """Return a new message type string with the specified version."""
        return MessageTypeStr(str(self._parsed.with_version(version)))



from unittest import IsolatedAsyncioTestCase

from .....messaging.request_context import RequestContext
from .....tests import mock
from .....utils.testing import create_test_profile
from .. import controller as test_module


class TestActionMenuController(IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        self.profile = await create_test_profile()
        self.context = RequestContext(self.profile)

    async def test_controller(self):
        menu_service = mock.MagicMock(test_module.BaseMenuService, autospec=True)
        self.profile.context.injector.bind_instance(
            test_module.BaseMenuService, menu_service
        )

        controller = test_module.Controller("protocol")

        assert await controller.determine_roles(self.profile.context) == ["provider"]
